import logging
from typing import Dict

import httpx

from .exceptions import JelasticAPIException


class JelasticAPIConnector:
    def __init__(self, apiurl: str, apitoken: str):
        """
        Get all needed data to connect to a Jelastic API
        """
        self.apiurl = apiurl
        self.apitoken = apitoken
        self.apidata = {"session": apitoken}
        self.logger = logging.getLogger(self.__class__.__name__)
        # httpx client. Default to no timeouts, as the Jelastic API is _synchronous_.
        self.client = httpx.Client(timeout=None)

    def is_functional(self) -> bool:
        """
        Whether this was meaningfully instantiated
        """
        try:
            return len(self.apiurl) > 0 and len(self.apidata["session"]) > 0
        except (TypeError, AttributeError):
            return False

    def _apicall(self, uri: str, method: str = "get", data: dict = {}) -> Dict:
        """
        Lowest-level API call: that's the method that talks over the network to the Jelastic API
        """
        # Make sure we have our session in
        self.logger.debug("_apicall {} {}, data:{}".format(method.upper(), uri, data))
        data.update(self.apidata)
        r = self.client.request(
            method=method, url="{url}{uri}".format(url=self.apiurl, uri=uri), data=data
        )
        if r.status_code != httpx.codes.OK:
            raise JelasticAPIException(
                "{method} to {uri} failed with HTTP code {code}".format(
                    method=method, uri=uri, code=r.status_code
                )
            )

        response = r.json()
        if response["result"] != 0:
            raise JelasticAPIException(
                "{method} to {uri} returned non-zero result: {result}".format(
                    method=method, uri=uri, result=response
                )
            )
        self.logger.debug(" response : {}".format(response))
        return response

    def _(self, function: str, **kwargs) -> Dict:
        """
        Direct API call, converting function paths into URLs; allows:
            JelasticAPIConnector._('Environment.Control.GetEnvs')
        """
        self.logger.info("{fnc}({data})".format(fnc=function, data=kwargs))
        # Determine function endpoint from the two-dotted string
        uri_chunks = function.split(".")
        if len(uri_chunks) != 3:
            raise JelasticAPIException(
                "Function ({fnc}) doesn't match standard Jelastic function (Group.Class.Function)".format(
                    fnc=function
                )
            )
        uri = "{grp}/{cls}/REST/{fnc}".format(
            grp=uri_chunks[0], cls=uri_chunks[1], fnc=uri_chunks[2]
        ).lower()

        return self._apicall(uri=uri, method="post", data=kwargs)



import warnings
from unittest.mock import Mock

import pytest

from jelapi import api_connector as jelapic
from jelapi.classes import JelasticEnvironment, JelasticNode, JelasticNodeGroup
from jelapi.exceptions import JelasticObjectException
from jelapi.factories import JelasticEnvironmentFactory

from .utils import get_standard_env, get_standard_node, get_standard_node_groups


def test_JelasticEnvironment_simple():
    """
    JelasticEnvironment can be instantiated as is
    """
    JelasticEnvironment()


def test_JelasticEnvironment_factory():
    """
    JelasticEnvironment can be instantiated as is
    """
    j = JelasticEnvironmentFactory()
    assert len(j.nodeGroups) == 3
    assert "cp" in j.nodeGroups.keys()
    assert "sqldb" in j.nodeGroups.keys()
    assert "storage" in j.nodeGroups.keys()


def test_JelasticEnvironment_deprecated():
    """
    JelasticEnvironment can be instantiated
    """
    with warnings.catch_warnings(record=True) as warns:
        JelasticEnvironment(jelastic_env=get_standard_env())
        assert len(warns) == 1


def test_JelasticEnvironment_ordering():
    """
    JelasticEnvironment can be instantiated, but the updates cannot be called in any order
    """
    j = JelasticEnvironment()
    with pytest.raises(JelasticObjectException):
        j.update_node_groups_from_info([])

    with pytest.raises(JelasticObjectException):
        j.update_nodes_from_info([])

    # Calling the update_from_env_dict first solves this
    j.update_from_env_dict(get_standard_env())
    j.update_node_groups_from_info([])
    j.update_nodes_from_info([])


def test_JelasticEnvironment_with_missing_data():
    """
    JelasticEnvironment cannot be instantiated with partial envInfo
    """
    env_truncated = get_standard_env()
    del env_truncated["domain"]

    # Deprecated format
    with warnings.catch_warnings(record=True):
        with pytest.raises(KeyError):
            JelasticEnvironment(jelastic_env=env_truncated)

    j = JelasticEnvironment()
    with pytest.raises(KeyError):
        j.update_from_env_dict(env_truncated)


def test_JelasticEnvironment_getter_by_name():
    """
    JelasticEnvironment.get() works, and does one call to api
    """
    jelapic()._ = Mock(
        return_value={"env": get_standard_env(), "envGroups": []},
    )
    assert isinstance(JelasticEnvironment.get("envName"), JelasticEnvironment)
    jelapic()._.assert_called_once()


def test_JelasticEnvironment_list_all():
    """
    JelasticEnvironment.get() works, and does one call to api
    """
    jelapic()._ = Mock(
        return_value={
            "infos": [
                {"env": get_standard_env(), "envGroups": []},
            ]
        },
    )
    JelasticEnvironment.list.cache_clear()
    jelenvs = JelasticEnvironment.list()
    assert isinstance(jelenvs, dict)
    first_jelenvname = list(jelenvs)[0]
    assert isinstance(jelenvs[first_jelenvname], JelasticEnvironment)
    jelapic()._.assert_called_once()

    # If we gather the list again, it will not get called more, thanks to the lru_cache:
    jelapic()._.reset_mock()
    JelasticEnvironment.list()
    jelapic()._.assert_not_called()

    # Let's clear the lru_cache
    JelasticEnvironment.list.cache_clear()
    JelasticEnvironment.list()
    jelapic()._.assert_called_once()


def test_JelasticEnvironment_list_with_nodes():
    """
    JelasticEnvironment can be instantiated with nodes
    """
    # With an empty node_groups, that'll fail.
    jelapic()._ = Mock(
        return_value={
            "infos": [
                {
                    "env": get_standard_env(),
                    "nodeGroups": [],
                    "nodes": [get_standard_node()],
                    "envGroups": [],
                },
            ]
        },
    )
    with pytest.raises(JelasticObjectException):
        JelasticEnvironment.list.cache_clear()
        jelenvs = JelasticEnvironment.list()

    jelapic()._ = Mock(
        return_value={
            "infos": [
                {
                    "env": get_standard_env(),
                    "nodeGroups": get_standard_node_groups(),
                    "nodes": [get_standard_node()],
                    "envGroups": [],
                },
            ]
        },
    )
    JelasticEnvironment.list.cache_clear()
    jelenvs = JelasticEnvironment.list()
    assert isinstance(jelenvs, dict)
    first_jelenvname = list(jelenvs)[0]
    assert isinstance(jelenvs[first_jelenvname], JelasticEnvironment)
    jelapic()._.assert_called_once()


def test_JelasticEnvironment_cannot_set_some_ro_attributes():
    """
    JelasticEnvironment can be instantiated, but some read-only attributes can be read, but not written
    """
    jelenv = JelasticEnvironmentFactory()
    for attr in ["shortdomain", "domain", "envName"]:
        assert getattr(jelenv, attr)
        with pytest.raises(AttributeError):
            setattr(jelenv, attr, "some arbitrary value not in the object")


def test_JelasticEnvironment_doesnt_differ_from_api_initially():
    """
    JelasticEnvironment can be instantiated, but some read-only attributes can be read, but not written
    """
    jelenv = JelasticEnvironmentFactory()
    assert not jelenv.differs_from_api()


def test_JelasticEnvironment_str_rep():
    """
    JelasticEnvironment can be instantiated, but some read-only attributes can be read, but not written
    """
    jelenv = JelasticEnvironmentFactory()
    assert str(jelenv) == "JelasticEnvironment 'envName' <https://domain>"


def test_JelasticEnvironment_differs_from_api_if_displayName_is_changed():
    """
    JelasticEnvironment can be instantiated, but some read-only attributes can be read, but not written
    """
    jelenv = JelasticEnvironmentFactory()
    jelenv.displayName = "different displayName"
    assert jelenv.differs_from_api()


def test_JelasticEnvironment_displayName_change_and_save_will_talk_to_API_twice():
    """
    JelasticEnvironment can be instantiated, but some read-only attributes can be read, but not written
    """

    jelapic()._ = Mock(
        return_value={"env": get_standard_env(), "envGroups": []},
    )

    jelenv = JelasticEnvironmentFactory()
    jelenv.displayName = "different displayName"
    jelenv._save_displayName()
    jelapic()._.assert_called()

    jelapic()._.reset_mock()

    # A second save should not call the API
    jelenv._save_displayName()
    jelapic()._.assert_not_called()


def test_JelasticEnvironment_differs_from_api_if_envGroups_is_changed():
    """
    JelasticEnvironment can be instantiated, but some read-only attributes can be read, but not written
    """
    jelenv = JelasticEnvironment()
    jelenv.update_from_env_dict(get_standard_env())
    jelenv.update_env_groups_from_info(["A", "B"])

    jelenv.envGroups.append("C")
    assert jelenv.differs_from_api()
    jelenv.envGroups = ["A", "B"]
    assert not jelenv.differs_from_api()
    jelenv.envGroups.remove("A")
    assert jelenv.differs_from_api()


def test_JelasticEnvironment_envGroups_change_and_save_will_talk_to_API():
    """
    JelasticEnvironment can be instantiated, but some read-only attributes can be read, but not written
    """
    jelapic()._ = Mock(
        return_value={"env": get_standard_env(), "envGroups": []},
    )

    jelenv = JelasticEnvironment(
        jelastic_env=get_standard_env(),
        env_groups=[
            "A",
            "B",
        ],
    )
    jelenv.envGroups.append("C")
    jelenv.save()
    jelapic()._.assert_called()

    jelapic()._.reset_mock()

    # A second save should not call the API
    jelenv.save()
    jelapic()._.assert_not_called()


def test_JelasticEnvironment_can_only_be_stopped_from_running():
    """
    JelasticEnvironment cannot (yet) be put to certain states
    """
    jelenv = JelasticEnvironment(
        jelastic_env=get_standard_env(),
        env_groups=[],
    )
    JelStatus = JelasticEnvironment.Status
    for status in [
        JelStatus.UNKNOWN,
        JelStatus.LAUNCHING,
        JelStatus.SUSPENDED,
        JelStatus.CREATING,
        JelStatus.CLONING,
        JelStatus.UPDATING,
    ]:

        jelenv._status = jelenv._from_api["status"] = status
        with pytest.raises(JelasticObjectException):
            jelenv.stop()

    # But it works from running
    jelapic()._ = Mock()
    jelenv._from_api["status"] = JelStatus.RUNNING
    jelenv._status = jelenv._from_api["status"] = JelStatus.RUNNING
    jelenv.stop()
    jelapic()._.assert_called_once()


def test_JelasticEnvironment_unsupported_statuses():
    """
    JelasticEnvironment cannot (yet) be put to certain states
    """
    jelenv = JelasticEnvironment(
        jelastic_env=get_standard_env(),
        env_groups=[],
    )
    JelStatus = JelasticEnvironment.Status
    for status in [
        JelStatus.UNKNOWN,
        JelStatus.LAUNCHING,
        JelStatus.SUSPENDED,
        JelStatus.CREATING,
        JelStatus.CLONING,
        JelStatus.UPDATING,
    ]:
        jelenv.status = status
        with pytest.raises(JelasticObjectException):
            jelenv.save()


def test_JelasticEnvironment_stop_via_status():
    """
    JelasticEnvironment can be started if started, by setting the status to STOPPED, and saving
    """
    jelapic()._ = Mock(
        return_value={
            "env": get_standard_env(
                status=JelasticEnvironment.Status.STOPPED.value
            ),  # After the stop, the API returns that it was stopped
            "envGroups": [],
        }
    )

    jelenv = JelasticEnvironment(
        jelastic_env=get_standard_env(),
        env_groups=[],
    )
    jelenv.status = JelasticEnvironment.Status.STOPPED
    jelenv.save()
    assert jelenv.status == JelasticEnvironment.Status.STOPPED
    jelapic()._.assert_called()

    # A second save should not call the API
    jelapic()._.reset_mock()
    jelenv.save()
    jelapic()._.assert_not_called()


def test_JelasticEnvironment_stop_via_method():
    """
    JelasticEnvironment can be stopped if running by running the stop() method
    """
    jelapic()._ = Mock(
        return_value={
            "env": get_standard_env(
                status=JelasticEnvironment.Status.STOPPED.value
            ),  # After the stop, the API returns that it was stopped
            "envGroups": [],
        }
    )

    jelenv = JelasticEnvironment(
        jelastic_env=get_standard_env(),
        env_groups=[],
    )
    jelenv.stop()
    assert jelenv.status == JelasticEnvironment.Status.STOPPED
    jelapic()._.assert_called()

    # A second save should not call the API
    jelapic()._.reset_mock()
    jelenv.save()
    jelapic()._.assert_not_called()


def test_JelasticEnvironment_start_via_status():
    """
    JelasticEnvironment can be started if stopped or sleeping, by setting the status to RUNNING, and saving
    """
    jelapic()._ = Mock(
        return_value={
            "env": get_standard_env(
                status=JelasticEnvironment.Status.RUNNING.value
            ),  # After the stop, the API returns that it was stopped
            "envGroups": [],
        }
    )

    # Test these two starting statuses
    for status in [
        JelasticEnvironment.Status.STOPPED,
        JelasticEnvironment.Status.SLEEPING,
    ]:
        jelapic()._.reset_mock()
        jelenv = JelasticEnvironment(
            jelastic_env=get_standard_env(status.value),
            env_groups=[],
        )
        jelenv.status = JelasticEnvironment.Status.RUNNING
        jelenv.save()
        assert jelenv.status == JelasticEnvironment.Status.RUNNING
        jelapic()._.assert_called()

        # A second save should not call the API
        jelapic()._.reset_mock()
        jelenv.save()
        jelapic()._.assert_not_called()


def test_JelasticEnvironment_start_via_method():
    """
    JelasticEnvironment can be started if stopped or sleeping, with the start() method
    """
    jelapic()._ = Mock(
        return_value={
            "env": get_standard_env(
                status=JelasticEnvironment.Status.RUNNING.value
            ),  # After the stop, the API returns that it was stopped
            "envGroups": [],
        }
    )

    # Test these two starting statuses
    for status in [
        JelasticEnvironment.Status.STOPPED,
        JelasticEnvironment.Status.SLEEPING,
    ]:
        jelapic()._.reset_mock()
        jelenv = JelasticEnvironment(
            jelastic_env=get_standard_env(status.value),
            env_groups=[],
        )
        jelenv.start()
        assert jelenv.status == JelasticEnvironment.Status.RUNNING
        jelapic()._.assert_called()

        # A second save should not call the API
        jelapic()._.reset_mock()
        jelenv.save()
        jelapic()._.assert_not_called()


def test_JelasticEnvironment_sleep_via_status():
    """
    JelasticEnvironment can be put to sleep if running, by setting the status to SLEEPING, and saving
    """
    jelapic()._ = Mock(
        return_value={
            "env": get_standard_env(
                status=JelasticEnvironment.Status.SLEEPING.value
            ),  # After the sleep, the API returns that it was sleeping
            "envGroups": [],
        }
    )

    # Test these two starting statuses
    for status in [
        JelasticEnvironment.Status.STOPPED,
        JelasticEnvironment.Status.SLEEPING,
    ]:
        jelapic()._.reset_mock()
        jelenv = JelasticEnvironment(
            jelastic_env=get_standard_env(status.value),
            env_groups=[],
        )
        jelenv.status = JelasticEnvironment.Status.RUNNING
        jelenv.save()
        assert jelenv.status == JelasticEnvironment.Status.SLEEPING
        jelapic()._.assert_called()

        # A second save should not call the API
        jelapic()._.reset_mock()
        jelenv.save()
        jelapic()._.assert_not_called()


def test_JelasticEnvironment_sleep_via_method():
    """
    JelasticEnvironment can be started if stopped or sleeping, with the start() method
    """
    jelapic()._ = Mock(
        return_value={
            "env": get_standard_env(
                status=JelasticEnvironment.Status.SLEEPING.value
            ),  # After the stop, the API returns that it was stopped
            "envGroups": [],
        }
    )

    jelenv = JelasticEnvironment(
        jelastic_env=get_standard_env(JelasticEnvironment.Status.RUNNING),
        env_groups=[],
    )
    jelenv.sleep()
    assert jelenv.status == JelasticEnvironment.Status.SLEEPING
    jelapic()._.assert_called()

    # A second save should not call the API
    jelapic()._.reset_mock()
    jelenv.save()
    jelapic()._.assert_not_called()


def test_JelasticEnvironment_differs_from_api_if_extdomains_is_changed():
    """
    JelasticEnvironment can be instantiated, but some read-only attributes can be read, but not written
    """
    jelenv = JelasticEnvironmentFactory()
    assert not jelenv.differs_from_api()
    jelenv.extdomains.append("test.example.com")
    assert jelenv.differs_from_api()
    jelenv.extdomains.append("test.example.org")
    assert jelenv.differs_from_api()


def test_JelasticEnvironment_extomains_change_and_save_will_talk_to_API():
    """
    JelasticEnvironment can be instantiated, but some read-only attributes can be read, but not written
    """
    twodomains = ["test.example.com", "test.example.org"]
    jelapic()._ = Mock(
        return_value={"env": get_standard_env(extdomains=twodomains), "envGroups": []},
    )

    jelenv = JelasticEnvironment(
        jelastic_env=get_standard_env(),
        env_groups=[],
    )
    jelenv.extdomains = twodomains
    jelenv.save()
    jelapic()._.assert_called()

    jelapic()._.reset_mock()

    #  Removing a domain also calls
    jelenv.extdomains.remove("test.example.com")
    jelenv.save()
    jelapic()._.assert_called()

    jelapic()._.reset_mock()

    # A second save should not call the API
    jelenv.save()
    jelapic()._.assert_not_called()


def test_JelasticEnvironment_nodes():
    """
    JelasticEnvironment can be instantiated with nodes
    """
    nodes = []
    for i in range(3):
        node = get_standard_node()
        node["id"] = i
        nodes.append(node)

    jelenv = JelasticEnvironment(
        jelastic_env=get_standard_env(),
        node_groups=get_standard_node_groups(),
        nodes=nodes,
    )
    assert not jelenv.differs_from_api()
    jelenv.nodeGroups["cp"].nodes[0].fixedCloudlets = 8
    assert jelenv.differs_from_api()

    jelapic()._ = Mock(
        return_value={
            "env": get_standard_env(),
            "envGroups": [],
            "nodes": [get_standard_node()],
        },
    )
    jelenv.save()
    assert not jelenv.differs_from_api()


def test_JelasticEnvironment_stranger_nodes():
    """
    JelasticEnvironment cannot be instantiated with nodes outside of the nodeGroups
    """
    nodes = [
        get_standard_node(),
    ]

    #  Get a node groups' list without above nodes' nodegroup
    node_groups = [
        ngdict
        for ngdict in get_standard_node_groups()
        if ngdict["name"] != nodes[0]["nodeGroup"]
    ]

    get_standard_node_groups()
    with pytest.raises(JelasticObjectException):
        JelasticEnvironment(
            jelastic_env=get_standard_env(),
            node_groups=node_groups,
            nodes=nodes,
        )


def test_JelasticEnvironment_node_fetcher():
    """
    Test the convennience node fetcher
    """
    nodes = []
    for i in range(2):
        node = get_standard_node()
        node["id"] = i
        nodes.append(node)

    nodes[0]["nodeGroup"] = "cp"
    nodes[1]["nodeGroup"] = "sqldb"

    jelenv = JelasticEnvironment(
        jelastic_env=get_standard_env(),
        node_groups=get_standard_node_groups(),
        nodes=nodes,
    )

    from jelapi.classes import JelasticNodeGroup

    # Do not look by full NodeGroup object
    with pytest.raises(JelasticObjectException):
        jelenv.node_by_node_group(JelasticNodeGroup.NodeGroupType.CACHE)

    # If the node's not around, exception
    with pytest.raises(JelasticObjectException):
        jelenv.node_by_node_group("nosqldb")

    assert isinstance(jelenv.node_by_node_group("cp"), JelasticNode)


def test_JelasticEnvironment_sumstats():
    """
    We can get Environment sumstats
    """
    jelapic()._ = Mock(
        return_value={"stats": []},  # Of course there is something in that dict.
    )
    jelenv = JelasticEnvironmentFactory()
    with pytest.raises(TypeError):
        # duration is needed
        jelenv.get_sumstats()

    # Fetch 8 hours'
    jelenv.get_sumstats(8 * 60 * 60)


def test_JelasticEnvironment_no_nodeGroups_wipe():
    """
    nodeGroups cannot be wiped
    """
    j = JelasticEnvironmentFactory()
    j.nodeGroups = {}
    assert j.differs_from_api()
    # We cannot wipe nodeGroups
    with pytest.raises(JelasticObjectException):
        j.save()


def test_JelasticEnvironment_add_node_group():
    """
    Test saving of nodeGroups' updates, adding one
    """
    j = JelasticEnvironmentFactory()
    ng = JelasticNodeGroup(nodeGroupType=JelasticNodeGroup.NodeGroupType.NOSQL_DATABASE)
    assert not ng.is_from_api
    assert ng._envVars == {}
    ng.attach_to_environment(j)
    ng.raise_unless_can_call_api()

    n = JelasticNode(nodeType=JelasticNode.NodeType.DOCKER)
    assert not n.is_from_api
    n.attach_to_node_group(ng)

    assert j.differs_from_api()

    jelapic()._ = Mock(
        return_value={
            "response": {
                "env": get_standard_env(),
                "envGroups": [],
                "nodes": [get_standard_node()],
            }
        },
    )
    j._save_topology_and_node_groups()
    # Called twice, once for saving, once for refresh
    jelapic()._.assert_called()


def test_JelasticEnvironment_add_node_group_error_in_api():
    """
    Test saving of nodeGroups', but it gives an error in API
    """
    j = JelasticEnvironmentFactory()
    ng = JelasticNodeGroup(nodeGroupType=JelasticNodeGroup.NodeGroupType.NOSQL_DATABASE)
    ng.attach_to_environment(j)
    ng.raise_unless_can_call_api()
    n = JelasticNode(nodeType=JelasticNode.NodeType.DOCKER)
    n.attach_to_node_group(ng)
    jelapic()._ = Mock(
        return_value={
            "response": {
                "error": "This is a long error message",
            }
        },
    )
    with pytest.raises(JelasticObjectException):
        j._save_topology_and_node_groups()
    # Called once, then rose
    jelapic()._.assert_called_once()


def test_JelEnv_cannot_clone_too_long():
    j = JelasticEnvironmentFactory()
    with pytest.raises(JelasticObjectException):
        # 34 chars is too long
        j.clone("abcdefghijklmnopqrstuvwxyz01234567")


def test_JelEnv_can_clone():
    j = JelasticEnvironmentFactory()
    # 33 chars is OK
    jelapic()._ = Mock(
        return_value={
            "env": get_standard_env(),
            "envGroups": [],
            "nodes": [get_standard_node()],
            "nodeGroups": get_standard_node_groups(),
        },
    )
    j.clone("abcdefghijklmnopqrstuvwxyz0123456")
    # Called twice actually
    jelapic()._.assert_called()



from datetime import datetime

import pytest

from jelapi.classes.jelasticobject import (
    _JelasticAttribute,
    _JelasticObject,
    _JelAttrBool,
    _JelAttrDatetime,
    _JelAttrDict,
    _JelAttrInt,
    _JelAttrIPv4,
    _JelAttrList,
    _JelAttrStr,
)


def test_JelasticAttribute_is_permissive_by_default():
    """
    _JelasticAttribute is a descriptor class
    """

    class Test:
        jela = _JelasticAttribute()

    t = Test()
    t.jela = "another string"
    t.jela = 2
    t.jela = [2, "string"]


def test_JelAttrBool_supports_bool_typecheck():
    """
    _JelAttrBool to store bools
    """

    class Test:
        jela = _JelAttrBool()

    t = Test()
    t.jela = False
    t.jela = True
    with pytest.raises(TypeError):
        t.jela = 2
    with pytest.raises(TypeError):
        t.jela = "string"


def test_JelAttrDatetime_supports_datetime_typecheck():
    """
    _JelAttrDatetimes to store datetimes
    """

    class Test:
        jela = _JelAttrDatetime()

    t = Test()
    t.jela = datetime.now()
    with pytest.raises(TypeError):
        t.jela = 2
    with pytest.raises(TypeError):
        t.jela = "string"
    with pytest.raises(TypeError):
        t.jela = False


def test_JelAttrStr_supports_str_typecheck():
    """
    _JelAttrStr to store strings
    """

    class Test:
        jela = _JelAttrStr()

    t = Test()
    t.jela = "a string"
    t.jela = "another string"
    with pytest.raises(TypeError):
        t.jela = 2
    with pytest.raises(TypeError):
        t.jela = [2, "string"]


def test_JelAttrIPv4_supports_ip_typecheck():
    """
    _JelAttrStr to store strings
    """

    class Test:
        ip = _JelAttrIPv4()

    t = Test()
    t.ip = "192.0.2.1"
    with pytest.raises(TypeError):
        # Too long
        t.ip = "192.0.2.1.2"
    with pytest.raises(TypeError):
        # Too short
        t.ip = "192.0.2"
    with pytest.raises(TypeError):
        # Out-of-range
        t.ip = "-1.0.2.1"
    with pytest.raises(TypeError):
        # Out-of-range
        t.ip = "256.0.2.1"
    with pytest.raises(TypeError):
        # NaN
        t.ip = "A.0.2.1"


def test_JelAttrInt_supports_int_typecheck():
    """
    _JelAttrInt to store ints
    """

    class Test:
        jela = _JelAttrInt()

    t = Test()
    t.jela = 3
    t.jela = 5
    with pytest.raises(TypeError):
        t.jela = "2"
    with pytest.raises(TypeError):
        t.jela = [2, "string"]


def test_JelAttrList_supports_list_typecheck():
    """
    _JelAttrList to store arbitrary lists
    """

    class Test:
        jela = _JelAttrList()

    t = Test()
    t.jela = ["2", "3"]
    t.jela.append("string")
    with pytest.raises(TypeError):
        t.jela = "2"
    with pytest.raises(TypeError):
        t.jela = 2


def test_JelAttrDict_supports_dict_typecheck():
    """
    _JelAttrDict to store arbitrary dicts
    """

    class Test:
        jela = _JelAttrDict()

    t = Test()
    t.jela = {"2": 3, "3": "text"}
    t.jela["1"] = "1"
    with pytest.raises(TypeError):
        t.jela = "2"
    with pytest.raises(TypeError):
        t.jela = 2


def test_JelasticAttribute_can_be_read_only():
    """
    _JelasticAttribute is a descriptor
    """

    class Test:
        jela = _JelasticAttribute(read_only=True)

    t = Test()
    # To set in nevertheless, set the private counterpart.
    t._jela = "a string"
    assert t.jela == "a string"
    with pytest.raises(AttributeError):
        t.jela = "another string"


def test_JelasticObject_is_abstract():
    """
    _JelasticObject is abstract, it can't be instantiated
    """
    with pytest.raises(TypeError):
        _JelasticObject()



from jelapi.classes import JelasticEnvironment, JelasticNodeGroup


def get_standard_node_group(
    node_group_type: JelasticNodeGroup.NodeGroupType = JelasticNodeGroup.NodeGroupType.APPLICATION_SERVER,
):
    return {"name": node_group_type.value}


def get_standard_node_groups():
    ngs = []
    for ngtype in JelasticNodeGroup.NodeGroupType:
        ngs.append({"name": ngtype.value})
    return ngs


def get_standard_env(status=JelasticEnvironment.Status.RUNNING.value, extdomains=None):
    extdomains = extdomains if extdomains else []
    return {
        "shortdomain": "shortdomain",
        "domain": "domain",
        "envName": "envName",
        "displayName": "initial displayName",
        "status": status,
        "extdomains": extdomains,
        "nodeGroups": get_standard_node_groups(),
        "hardwareNodeGroup": "a_provider_specific_string",
        "sslstate": True,
        "ishaenabled": False,
    }


def get_standard_node(
    id: int = 987, fixed_cloudlets: int = 1, flexible_cloudlets: int = 1
):
    return {
        "id": id,
        "fixedCloudlets": fixed_cloudlets,
        "flexibleCloudlets": flexible_cloudlets,
        "intIP": "192.0.2.1",
        "nodeGroup": "cp",
        "url": "https://test.example.com",
        "diskIoLimit": 100000,
        "diskIopsLimit": 1000,
        "diskLimit": 20000,
        "endpoints": [],
        "features": ["FIREWALL"],
        "hasPackages": False,
        "isClusterSupport": False,
        "isCustomSslSupport": False,
        "isExternalIpRequired": False,
        "isHighAvailability": False,
        "isResetPassword": False,
        "isVcsSupport": False,
        "isWebAccess": True,
        "ismaster": True,
        "maxchanks": 32,
        "messages": [],
        "name": "image-name",
        "nodeGroup": "cp",
        "nodeType": "docker",
        "nodemission": "docker",
        "osType": "LINUX",
        "packages": [],
        "port": 22,
        "singleContext": False,
        "status": 1,
        "type": "NATIVE",
        "version": "docker-tag",
    }


def get_standard_mount_point(source_node_id: int = 1):
    return {
        "name": "mount point name",
        "path": "/tmp/test",
        "sourcePath": "/tmp/sourcePath",
        "sourceNodeId": source_node_id,
    }



from copy import deepcopy
from enum import Enum
from functools import lru_cache
from json import dumps as jsondumps
from typing import Any, Dict, List, Optional

from ..exceptions import JelasticObjectException, deprecation
from .jelasticobject import _JelasticAttribute as _JelAttr
from .jelasticobject import (
    _JelasticObject,
    _JelAttrBool,
    _JelAttrDict,
    _JelAttrList,
    _JelAttrStr,
)
from .node import JelasticNode
from .nodegroup import JelasticNodeGroup


class JelasticEnvironment(_JelasticObject):
    """
    Represents a Jelastic Environment
    """

    class Status(Enum):
        """
        Standard statuses
        Latest reference https://github.com/jelastic/localizations/blob/master/English/5.4/%5Blang-en.js%5D#L999
        """

        UNKNOWN = 0
        RUNNING = 1
        STOPPED = 2
        LAUNCHING = 3
        SLEEPING = 4
        SUSPENDED = 5
        CREATING = 6
        CLONING = 7
        EXPORTING = 9
        MIGRATING = 10
        BROKEN = 11
        UPDATING = 12
        STOPPING = 13
        GOING_TO_SLEEP = 14
        REFRESHING = 1002

    displayName = _JelAttrStr()
    envGroups = _JelAttrList()
    status = _JelAttr()
    envName = _JelAttrStr(read_only=True)
    shortdomain = _JelAttrStr(read_only=True)
    domain = _JelAttrStr(read_only=True)
    extdomains = _JelAttrList()
    nodeGroups = _JelAttrDict(checked_for_differences=False)
    ishaneabled = _JelAttrBool(read_only=True)
    hardwareNodeGroup = _JelAttrStr(read_only=True)
    sslstate = _JelAttrBool()

    @staticmethod
    def get(envName: str) -> "JelasticEnvironment":
        """
        Static method to get one environment
        """
        # This is needed as it's a static method
        from .. import api_connector as jelapi_connector

        response = jelapi_connector()._(
            "Environment.Control.GetEnvInfo", envName=envName
        )
        j = JelasticEnvironment()
        j.update_from_env_dict(response["env"])
        j.update_env_groups_from_info(response.get("envGroups", []))
        j.update_node_groups_from_info(response.get("nodeGroups", []))
        j.update_nodes_from_info(response.get("nodes", []))
        return j

    @staticmethod
    @lru_cache(maxsize=1)
    def list() -> Dict[str, "JelasticEnvironment"]:
        """
        Static method to get all environments
        """
        # This is needed as it's a static method
        from .. import api_connector as jelapi_connector

        response = jelapi_connector()._("Environment.Control.GetEnvs")
        envs = {}
        for info in response["infos"]:
            name = info["env"]["envName"]
            envs[name] = JelasticEnvironment()
            envs[name].update_from_env_dict(info["env"])
            envs[name].update_env_groups_from_info(info.get("envGroups", []))
            envs[name].update_node_groups_from_info(info.get("nodeGroups", []))
            envs[name].update_nodes_from_info(info.get("nodes", []))

        return envs

    def clone(self, cloned_environment_name: str) -> "JelasticEnvironment":
        """
        Clone an environment, and return a new JelasticEnvironment matching the new one
        """
        if len(cloned_environment_name) > 33:
            raise JelasticObjectException(
                "New environments' names cannot be longer than 33 characters"
            )
        self.api._(
            "Environment.Control.CloneEnv",
            srcEnvName=self.envName,
            dstEnvName=cloned_environment_name,
        )
        return JelasticEnvironment.get(envName=cloned_environment_name)

    def attach_node_group(self, node_group: JelasticNodeGroup) -> None:
        """
        Make sure a node_group is attached correctly to that Environment
        """
        self.nodeGroups[node_group.nodeGroupType.value] = node_group
        node_group._parent = self

    def update_from_env_dict(self, jelastic_env_dict: Dict[str, Any]) -> None:
        """
        Update from the environment dict as gotten from API
        """
        # Allow exploration of the returned object, but don't act on it.
        self._env = jelastic_env_dict
        # Read-only attributes
        self._shortdomain = self._env["shortdomain"]
        self._envName = self._env["envName"]
        self._domain = self._env["domain"]
        self._hardwareNodeGroup = self._env["hardwareNodeGroup"]
        self._sslstate = self._env["sslstate"]
        self._ishaneabled = self._env["ishaenabled"]

        # Read-write attributes
        # displayName is sometimes not-present, do not die
        self.displayName = self._env.get("displayName", "")
        self.status = next(
            (status for status in self.Status if status.value == self._env["status"]),
            self.Status.UNKNOWN,
        )
        self.extdomains = self._env["extdomains"]

        # Copy our attributes as it came from API
        self.copy_self_as_from_api()

    def update_env_groups_from_info(self, env_groups: List[str]) -> None:
        """
        Update the envGroups as coming from API
        """
        self.envGroups = env_groups
        self.copy_self_as_from_api("envGroups")

    def update_node_groups_from_info(self, node_groups: List[Dict[str, Any]]) -> None:
        """
        Set the node groups (as gotten from API)
        """
        if not hasattr(self, "_envName"):
            raise JelasticObjectException(
                "update_node_groups: envName unset; call update_from_env_dict() first !"
            )

        for node_group_from_env in node_groups:
            node_group = JelasticNodeGroup()
            node_group.update_from_env_dict(node_group_from_env=node_group_from_env)
            node_group.attach_to_environment(self)

        self.copy_self_as_from_api("nodeGroups")

    def update_nodes_from_info(self, nodes: List[Dict[str, Any]]) -> None:
        """
        Construct/Update the nodes
        """
        if not hasattr(self, "_envName"):
            raise JelasticObjectException(
                "update_nodes: envName unset; call update_from_env_dict() first !"
            )

        for ng in self.nodeGroups.values():
            ng.nodes = []

        # Now add nodes in the nodeGroup
        for node_dict in nodes:
            if node_dict["nodeGroup"] not in self.nodeGroups:
                raise JelasticObjectException(
                    "Environment got a node outside of one of its nodeGroups"
                )

            node_group = self.nodeGroups[node_dict["nodeGroup"]]
            jelnode = JelasticNode()
            jelnode.update_from_env_dict(node_from_env=node_dict)
            jelnode.attach_to_node_group(node_group)

            node_group.copy_self_as_from_api("nodes")

    def __init__(
        self,
        *,
        jelastic_env: Dict[str, Any] = None,
        env_groups: Optional[List[str]] = None,
        node_groups: Optional[List[Dict[str, Any]]] = None,
        nodes: Optional[List[Dict[str, Any]]] = None,
    ) -> None:
        """
        Construct a JelasticEnvironment from various data sources
        """
        super().__init__()
        if not hasattr(self, "nodeGroup"):
            self.nodeGroups = {}
        if not hasattr(self, "envGroups"):
            self.envGroups = []

        if jelastic_env:
            deprecation(
                "JelasticEnvironment.__init__(): passing jelastic_env is deprecated; use update_from_env_dict() instead"
            )
            self.update_from_env_dict(jelastic_env)

        if env_groups:
            deprecation(
                "JelasticEnvironment.__init__(): passing env_groups is deprecated; use update_env_groups_from_info() instead"
            )

            self.update_env_groups_from_info(env_groups)

        if node_groups:
            deprecation(
                "JelasticEnvironment.__init__(): passing node_groups is deprecated; use update_node_groups_from_info() instead"
            )
            self.update_node_groups_from_info(node_groups)

        if nodes:
            deprecation(
                "JelasticEnvironment.__init__(): passing nodes is deprecated; use update_nodes_from_info() instead"
            )
            self.update_nodes_from_info(nodes)

    def refresh_from_api(self) -> None:
        response = self.api._("Environment.Control.GetEnvInfo", envName=self.envName)
        self.update_from_env_dict(response["env"])

        self.update_env_groups_from_info(response.get("envGroups", []))
        self.update_node_groups_from_info(response.get("nodeGroups", []))
        self.update_nodes_from_info(response.get("nodes", []))

    def __str__(self) -> str:
        return f"JelasticEnvironment '{self.envName}' <https://{self.domain}>"

    def _save_displayName(self):
        """
        Propagate the displayName change to the Jelastic API
        """
        if self.displayName != self._from_api["displayName"]:
            self.api._(
                "Environment.Control.SetEnvDisplayName",
                envName=self.envName,
                displayName=self.displayName,
            )
            self._from_api["displayName"] = self.displayName

    def _save_envGroups(self):
        """
        Propagate the envGroups change to the Jelastic API
        """
        if self.envGroups != self._from_api["envGroups"]:
            self.api._(
                "Environment.Control.SetEnvGroup",
                envName=self.envName,
                envGroups=jsondumps(self.envGroups),
            )
            self.copy_self_as_from_api("envGroups")

    def _save_extDomains(self):
        """
        Ensure the extDomains are bound correctly
        """
        if self.extdomains != self._from_api["extdomains"]:
            # Remove domains
            for domain in self._from_api["extdomains"]:
                if domain not in self.extdomains:
                    self.api._(
                        "Environment.Binder.RemoveExtDomain",
                        envName=self.envName,
                        extdomain=domain,
                    )
            # Add domains
            for domain in self.extdomains:
                if domain not in self._from_api["extdomains"]:
                    self.api._(
                        "Environment.Binder.BindExtDomain",
                        envName=self.envName,
                        extdomain=domain,
                    )
            self._from_api["extdomains"] = self.extdomains

    def _set_running_status(self, to_status_now: Status):
        """
        Put Environment in the right status
        """
        self.status = to_status_now

        if self.status != self._from_api["status"]:
            if self.status == self.Status.RUNNING:
                # TODO limit the statuses from which this is possible
                self.api._(
                    "Environment.Control.StartEnv",
                    envName=self.envName,
                )
                self._from_api["status"] = self.Status.RUNNING
            elif self.status == self.Status.STOPPED:
                if self._from_api["status"] not in [self.Status.RUNNING]:
                    raise JelasticObjectException(
                        "Cannot stop an environment not running"
                    )
                self.api._(
                    "Environment.Control.StopEnv",
                    envName=self.envName,
                )
                self._from_api["status"] = self.Status.STOPPED
            elif self.status == self.Status.SLEEPING:
                self.api._(
                    "Environment.Control.SleepEnv",
                    envName=self.envName,
                )
                self._from_api["status"] = self.Status.SLEEPING
            else:
                raise JelasticObjectException(
                    f"{self.__class__.__name__}: {self.status} not supported"
                )

    def get_topology(self) -> Dict[str, Any]:
        """
        Return the "Environment" topology, as consumed by ChangeTopology
        See https://docs.jelastic.com/api/5.9.8/public/#!/api/environment.Control-method-ChangeTopology
        """
        return {
            "displayName": self.displayName,
            "ishaenabled": self.ishaneabled,
            "region": self.hardwareNodeGroup,
            "shortdomain": self.shortdomain,
            "sslstate": self.sslstate,
        }
        # Missing keys:
        # "engine": "string",

    def _save_topology_and_node_groups(self):
        """
        Save the topology (nodeGroups, sslstate and others), then the nodeGroups'.
        """
        # Determine if a topology change is needed
        if (
            "nodeGroups" not in self._from_api
            or len(self._from_api["nodeGroups"]) != len(self.nodeGroups)
            or any(ng.needs_topology_update() for ng in self.nodeGroups.values())
            or self.sslstate != self._from_api["sslstate"]
        ):
            # We need to force the API to match what we want.
            # First, no wipeout of nodeGroups
            if len(self.nodeGroups) == 0:
                raise JelasticObjectException("Wipeout of nodeGroups not allowed")

            # Backup the stuff we want to keep after topology change
            wanted_node_groups = deepcopy(self.nodeGroups)

            apiresponse = self.api._(
                "Environment.Control.ChangeTopology",
                envName=self.envName,
                env=jsondumps(self.get_topology()),
                nodes=jsondumps([ng.get_topology() for ng in self.nodeGroups.values()]),
            )
            response = apiresponse["response"]

            if "error" in response:
                raise JelasticObjectException(
                    "There was an error in ChangeTopology: {}".format(response["error"])
                )

            self.update_from_env_dict(response["env"])
            self.update_env_groups_from_info(response.get("envGroups", []))
            self.nodeGroups = wanted_node_groups
            self.update_nodes_from_info(response.get("nodes", []))

            for k, ng in self.nodeGroups.items():
                for n in ng.nodes:
                    n.copy_self_as_from_api()

                # Complex ones
                ng.copy_self_as_from_api("_envVars")
                ng.copy_self_as_from_api("_links")
                ng.copy_self_as_from_api("_containerVolumes")

                # Bare attributes
                ng.copy_self_as_from_api("displayName")

                # Make sure these get saved afterwards
                ng._from_api["_mountPoints"] = []

        for ng in self.nodeGroups.values():
            ng.save()
        self._from_api["nodeGroups"] = self.nodeGroups

    def save_to_jelastic(self):
        """
        Mandatory _JelasticObject method, to save status to Jelastic
        """
        self._save_displayName()
        self._save_envGroups()
        self._save_extDomains()
        self._set_running_status(self.status)
        self._save_topology_and_node_groups()

    def node_by_node_group(self, node_group: str) -> JelasticNode:
        """
        Return a node by nodeGroup magic string
        """
        valid_node_groups = [ng.value for ng in JelasticNodeGroup.NodeGroupType]
        if node_group not in valid_node_groups:
            raise JelasticObjectException(
                f"node_group value {node_group} not in {valid_node_groups}"
            )

        try:
            return self.nodeGroups[node_group].nodes[0]
        except (IndexError, KeyError):
            raise JelasticObjectException(
                f"node_group {node_group} not found in environment's nodes"
            )

    def get_sumstats(self, duration_in_seconds: int) -> List[str]:
        """
        Get usage stats
        """
        response = self.api._(
            "Environment.Control.GetSumStat",
            envName=self.envName,
            duration=duration_in_seconds,
        )
        return response["stats"]

    def start(self) -> None:
        """
        Start Environment immediately
        """
        self._set_running_status(self.Status.RUNNING)

    def stop(self) -> None:
        """
        Stop Environment immediately
        """
        self._set_running_status(self.Status.STOPPED)

    def sleep(self) -> None:
        """
        Put Environment to sleep immediately
        """
        self._set_running_status(self.Status.SLEEPING)



import logging
from abc import ABC, abstractmethod
from copy import deepcopy
from datetime import datetime
from typing import Any, Dict, Optional


class _JelasticAttribute:
    """
    Descriptor class, with two tweakables:
    - read_only
    - checked_for_differences
    """

    def __init__(self, read_only: bool = False, checked_for_differences: bool = True):
        self.read_only = read_only
        self.checked_for_differences = checked_for_differences

    def __set_name__(self, owner, name):
        self.public_name = name
        self.private_name = f"_{name}"

    def __get__(self, obj: Any, objtype: type = None) -> Any:
        return getattr(obj, self.private_name)

    def __set__(self, obj: Any, value: Any):

        if self.read_only:
            raise AttributeError(
                f"{self.__class__.__name__}: '{self.public_name}' is read only."
            )
        self.typecheck(value)
        setattr(obj, self.private_name, value)

    def typecheck(self, value: Any) -> None:
        """
        :raises TypeError if the typecheck fails
        """
        pass


class _JelAttrStr(_JelasticAttribute):
    def typecheck(self, value: Any) -> None:
        if not isinstance(value, str):
            raise TypeError(f"{value} is no str")


class _JelAttrBool(_JelasticAttribute):
    def typecheck(self, value: Any) -> None:
        if not isinstance(value, bool):
            raise TypeError(f"{value} is no bool")


class _JelAttrInt(_JelasticAttribute):
    def typecheck(self, value: Any) -> None:
        if not isinstance(value, int):
            raise TypeError(f"{value} is no int")


class _JelAttrDatetime(_JelasticAttribute):
    def typecheck(self, value: Any) -> None:
        if not isinstance(value, datetime):
            raise TypeError(f"{value} is no datetime")


class _JelAttrDict(_JelasticAttribute):
    def typecheck(self, value: Any) -> None:
        if not isinstance(value, dict):
            raise TypeError(f"{value} is no dict")
        # Checking the dict _items_ for type doesn't work, see List


class _JelAttrList(_JelasticAttribute):
    def typecheck(self, value: Any) -> None:
        if not isinstance(value, list):
            raise TypeError(f"{value} is no list")
        # Checking the list _items_ for type doesn't work reliably; one can l.append(item) and it won't be checked


class _JelAttrIPv4(_JelAttrStr):
    def typecheck(self, value: Any) -> None:
        super().typecheck(value)
        ip_chunks = value.split(".")
        if len(ip_chunks) != 4:
            raise TypeError(f"{value} is no IPv4 address")
        for n in ip_chunks:
            try:
                if int(n) < 0 or int(n) > 255:
                    raise TypeError(f"{value} is no IPv4 address ({n} is out of range)")
            except ValueError:
                raise TypeError(f"{value} is no IPv4 address ({n} is no int)")


class _JelasticObject(ABC):
    """
    Any Jelastic Object, that keeps the last data as fetched from the API
    _from_api                dict of attributes as last refreshed from API
    """

    _from_api: Optional[Dict[str, Any]] = None
    _logger: logging.Logger

    def __init__(self, *args, **kwargs) -> None:
        """
        Instantiate logger
        """
        self._logger = logging.getLogger(self.__class__.__name__)

    def _tracelog(self, msg, *args, **kwargs):
        """
        Coding-level tracer, if needed
        """
        return self._logger.log(logging.DEBUG - 1, msg, *args, **kwargs)

    def __deepcopy__(self, memo):
        """
        When copying the object, mark it as not from the API
        """
        # See https://stackoverflow.com/questions/1500718/how-to-override-the-copy-deepcopy-operations-for-a-python-object
        cls = self.__class__
        cp = cls.__new__(cls)
        memo[id(self)] = cp
        for k, v in self.__dict__.items():
            setattr(cp, k, deepcopy(v, memo))

        cp._from_api = []
        return cp

    def archive_from_api(self):
        """
        Get a deepcopy of thyself, cut from API
        """
        self.before_archive_from_api()
        cp = deepcopy(self)
        cp.after_archive_from_api()
        return cp

    def before_archive_from_api(self):
        """
        Do what's needed after deepcopying away from API
        """

    def after_archive_from_api(self):
        """
        Do what's needed after deepcopying away from API
        """

    def copy_self_as_from_api(self, only_this_key: str = None) -> None:
        """
        Store a copy of ourselves, as it was from API
        """
        # Instantiate dict
        if not self._from_api:
            self._from_api = {}
        for k, v in vars(self).items():
            if k[0] == "_":
                # Check public_name
                k = k[1:]
            if only_this_key and k != only_this_key:
                # Only handle one key, and it's not this one
                continue
            # This convoluted syntax checks if self.k is a _JelasticAttribute
            # These are the only ones we want to copy
            try:
                descriptor_class = vars(type(self))[k]
            except KeyError:
                descriptor_class = None

            if (
                isinstance(descriptor_class, _JelasticAttribute)
                and not descriptor_class.read_only
            ):
                self._from_api[k] = deepcopy(v)

        self._from_api["copied_to_api_at"] = datetime.now()

    @property
    def is_from_api(self) -> bool:
        """
        Whether it was from API or is a new instance
        """
        try:
            return hasattr(self, "_from_api") and "copied_to_api_at" in self._from_api
        except TypeError:
            return False

    def differs_from_api(self) -> bool:
        """
        Check if the JelasticAttributes differ from the API
        """
        if not self.is_from_api:
            self._tracelog(f"differs_from_api() = {True} (as is_from_api = {False})")
            return True

        for k, v in vars(self).items():
            if k[0] == "_":
                # Check public_name
                k = k[1:]
            # This convoluted syntax checks if self.k is a _JelasticAttribute
            # These are the only ones we want to copy
            try:
                descriptor_class = vars(type(self))[k]
            except KeyError:
                descriptor_class = None
            if (
                isinstance(descriptor_class, _JelasticAttribute)
                and not descriptor_class.read_only
            ):
                if descriptor_class.checked_for_differences:
                    if k not in self._from_api or self._from_api[k] != v:
                        self._tracelog(
                            f"differs_from API because k:{k} was checked and differs"
                        )
                        return True
                elif isinstance(descriptor_class, _JelAttrList):
                    self._tracelog(f"Check if list {k} is in _from_api")
                    if k not in self._from_api:
                        self._tracelog(
                            f"differs_from API because {k} is not in _from_api"
                        )
                        return True

                    self._tracelog(
                        f"Check if list {k} differs from API; {v} vs {self._from_api[k]}"
                    )
                    if len(v) != len(self._from_api[k]):
                        self._tracelog(
                            f"differs_from API because list:{k} was checked for length and differs from API ({len(v)} != {len(self._from_api[k])})"
                        )
                        return True
                    if any(item.differs_from_api() for item in v):
                        self._tracelog(
                            f"differs_from API because list:{k} was checked and one item differs"
                        )
                        return True
                elif isinstance(descriptor_class, _JelAttrDict):
                    self._tracelog(
                        f"Check if dict {k} differs from API; {v} vs {self._from_api[k]}"
                    )
                    if len(v) != len(self._from_api[k]):
                        self._tracelog(
                            f"differs_from API because dict:{k} was checked for length and differs from API ({len(v)} != {len(self._from_api[k])})"
                        )
                        return True
                    if any(item.differs_from_api() for item in v.values()):
                        self._tracelog(
                            f"differs_from API because dict:{k} was checked and one item differs"
                        )
                        return True
        return False

    @abstractmethod
    def save_to_jelastic(self) -> None:
        """
        If needed, do what's needed to save the object to the API
        DO NOT update the object. That'd done in refresh_from_api
        """

    def save(self) -> None:
        """
        Save the changes staged in attributes
        """
        if self.differs_from_api():
            # Implements the saving of the changes to Jelastic
            self._tracelog("save() -> differs_from_api() -> save_to_jelastic()")
            self.save_to_jelastic()
            if hasattr(self, "refresh_from_api"):
                # Fetches, to verify changes were proceeded with correctly
                self._tracelog("save() -> differs_from_api() -> refresh_from_api()")
                self.refresh_from_api()

        # Make extra sure we did update everything needed, and that all sub saves behaved correctly
        assertmsg = f" {self.__class__.__name__}: save_to_jelastic() method only partially implemented."
        assert not self.differs_from_api(), assertmsg

    @property
    def api(self):
        """
        Return the global api connector, as property
        """
        from .. import api_connector as jelapi_connector

        return jelapi_connector()



from typing import Any, Dict

from ..exceptions import JelasticObjectException, deprecation
from ._volume import _JelasticVolume
from .jelasticobject import _JelasticAttribute as _JelAttr
from .jelasticobject import _JelAttrStr
from .node import JelasticNode
from .nodegroup import JelasticNodeGroup


class JelasticMountPoint(_JelasticVolume):
    """
    Represents a Jelastic MountPoint, a mount link between nodes
    """

    # Where it is mounted _from_
    sourceNode = _JelAttr(read_only=True)
    sourcePath = _JelAttrStr(read_only=True)

    def update_from_env_dict(self, mount_point_from_api: Dict[str, Any]) -> None:
        """
        Construct/Update our object from the structure
        """
        if not hasattr(self, "_nodeGroup"):
            raise JelasticObjectException(
                "MountPoint cannot be updated from dict without being attached to a nodeGroup first"
            )

        # Allow exploration of the returned object, but don't act on it.
        self._mount_point = mount_point_from_api

        # Set the name and path in _JelasticVolume
        self._name = self._mount_point["name"]
        self._path = self._mount_point["path"]

        self._sourcePath = self._mount_point["sourcePath"]

        # Now find internal source node
        source_node_id = int(self._mount_point["sourceNodeId"])

        for ng in self._nodeGroup._parent.nodeGroups.values():
            for node in ng.nodes:
                if node.id == source_node_id:
                    self._sourceNode = node

        if not hasattr(self, "_sourceNode"):
            raise JelasticObjectException(
                f"Didn't find node id {source_node_id} in environment"
            )

        # Copy our attributes as it came from API
        self.copy_self_as_from_api()

    def __init__(
        self,
        *,
        node_group: JelasticNodeGroup = None,
        mount_point_from_api: Dict[str, Any] = None,
        name: str = None,
        path: str = None,
        sourceNode: JelasticNode = None,
        sourcePath: str = None,
    ) -> None:
        """
        Construct a JelasticMountPoint from the outer data
        """
        super().__init__(name=name, path=path)

        if node_group:
            deprecation(
                "mountPoint.__init__(): passing node_group is deprecated; use attach_to_node_group() instead"
            )
            self.attach_to_node_group(node_group)

        if mount_point_from_api:
            if not node_group:
                raise TypeError(
                    "The deprecated __init__ call needs both node_group and mount_point_from_api"
                )
            deprecation(
                "mountPoint.__init__(): passing mount_point_from_api is deprecated; use update_from_env_dict() instead"
            )
            self.update_from_env_dict(mount_point_from_api=mount_point_from_api)
            assert not self.differs_from_api()

        if sourceNode and sourcePath:
            # These go in RO attributes
            self._sourceNode = sourceNode
            self._sourcePath = sourcePath
            assert not self.is_from_api
            assert self.differs_from_api()

    def add_to_api(self) -> None:
        """
        Push this mountpoint to API
        """
        if self.is_from_api:
            raise JelasticObjectException(
                "MountPoint cannot be added to API, it came from it"
            )
        # Needs to be added
        self.api._(
            "Environment.File.AddMountPointByGroup",
            envName=self._envName,
            nodeGroup=self._nodeGroup.nodeGroupType.value,
            path=self.path,
            sourceNodeId=self.sourceNode.id,
            sourcePath=self.sourcePath,
        )
        self.copy_self_as_from_api()
        assert not self.differs_from_api()

    def del_from_api(self) -> None:
        """
        Delete this mountpoint from API
        """
        if not self.is_from_api:
            raise JelasticObjectException(
                "MountPoint cannot be removed from API, as it did not come from it"
            )
        # Needs to be removed
        self.api._(
            "Environment.File.RemoveMountPointByGroup",
            envName=self._envName,
            nodeGroup=self._nodeGroup.nodeGroupType.value,
            path=self.path,
        )
        self.copy_self_as_from_api()
        assert not self.differs_from_api()

    def save_to_jelastic(self):
        """
        Mandatory _JelasticObject method, to save status to Jelastic
        """
        # TODO

    def __str__(self) -> str:
        """
        String representation
        """
        return f"JelasticMountPount {self.path} on {self.nodeGroup} - from {self.sourcePath} on {self.sourceNode.id}"



import json
from enum import Enum
from typing import Any, Dict, List

from ..exceptions import JelasticObjectException, deprecation
from .jelasticobject import _JelasticAttribute as _JelAttr
from .jelasticobject import (
    _JelasticObject,
    _JelAttrBool,
    _JelAttrInt,
    _JelAttrIPv4,
    _JelAttrList,
    _JelAttrStr,
)
from .nodegroup import JelasticNodeGroup


class JelasticNode(_JelasticObject):
    """
    Represents a Jelastic Node
    """

    class NodeType(Enum):
        """
        Available Node Types
        See https://docs.cloudscripting.com/creating-manifest/selecting-containers/#supported-stacks
        """

        DOCKER = "docker"
        STORAGE = "storage"

    id = _JelAttrInt(read_only=True)
    envName = _JelAttrStr(read_only=True)
    intIP = _JelAttrIPv4(read_only=True)
    url = _JelAttrStr(read_only=True)
    nodeGroup = _JelAttr(read_only=True)
    status = _JelAttr(read_only=True)

    # Not a real attribute, at least not synced to API
    docker_registry: Dict
    docker_image = _JelAttrStr()

    # TODO: make the ones that make sense as RW attributes
    diskIoLimit = _JelAttrInt(read_only=True)
    diskIopsLimit = _JelAttrInt(read_only=True)
    diskLimit = _JelAttrInt(read_only=True)
    endpoints = _JelAttrList(read_only=True)
    features = _JelAttrList(read_only=True)
    hasPackages = _JelAttrBool(read_only=True)
    isClusterSupport = _JelAttrBool(read_only=True)
    isCustomSslSupport = _JelAttrBool(read_only=True)
    isExternalIpRequired = _JelAttrBool(read_only=True)
    isHighAvailability = _JelAttrBool(read_only=True)
    isResetPassword = _JelAttrBool(read_only=True)
    isVcsSupport = _JelAttrBool(read_only=True)
    isWebAccess = _JelAttrBool(read_only=True)
    ismaster = _JelAttrBool(read_only=True)
    maxchanks = _JelAttrInt(read_only=True)
    messages = _JelAttrList(read_only=True)
    name = _JelAttrStr(read_only=True)
    nodeType = _JelAttr(read_only=True)
    nodemission = _JelAttrStr(read_only=True)
    osType = _JelAttrStr(read_only=True)
    packages = _JelAttrList(read_only=True)
    port = _JelAttrInt(read_only=True)
    singleContext = _JelAttrBool(read_only=True)
    type = _JelAttrStr(read_only=True)
    version = _JelAttrStr(read_only=True)

    # "always guaranteed minimum"
    fixedCloudlets = _JelAttrInt()
    # "maximum"
    flexibleCloudlets = _JelAttrInt()
    allowFlexibleCloudletsReduction = _JelAttrBool(checked_for_differences=False)

    def attach_to_node_group(self, node_group: "JelasticNodeGroup") -> None:
        """
        Set the nodeGroup, with all accompanying things
        """
        node_group.append_node(self)

        # Read-only attributes
        try:
            self._envName = self._nodeGroup._parent.envName
        except AttributeError:
            # Do not set it, it'll fail the self.raise_unless_can_update_to_api() check if unset
            pass

    def update_from_env_dict(
        self,
        node_from_env: Dict[str, Any],
    ) -> None:
        """
        Construct/Update our object from the structure
        """
        # Allow exploration of the returned object, but don't act on it.
        self._node = node_from_env

        self._nodeType = next(
            (nt for nt in self.NodeType if nt.value == self._node["nodeType"]), None
        )
        if not self.nodeType:
            raise JelasticObjectException(f"nodeType unknown: {self._node['nodeType']}")

        # Mandatory attributes, raises KeyError if missing
        for attr in [
            "id",
            "name",
            "nodemission",
            "type",
        ]:
            setattr(self, f"_{attr}", self._node[attr])

        from .environment import JelasticEnvironment

        self._status = next(
            (
                status
                for status in JelasticEnvironment.Status
                if status.value == self._node["status"]
            ),
            JelasticEnvironment.Status.UNKNOWN,
        )

        # Ususal attributes, does not raise if inexistant
        for attr in [
            "intIP",
            "url",
            "diskIoLimit",
            "diskIopsLimit",
            "diskLimit",
            "endpoints",
            "features",
            "hasPackages",
            "isClusterSupport",
            "isCustomSslSupport",
            "isExternalIpRequired",
            "isHighAvailability",
            "isResetPassword",
            "isVcsSupport",
            "isWebAccess",
            "ismaster",
            "maxchanks",
            "messages",
            "osType",
            "packages",
            "port",
            "singleContext",
            "url",
            "version",
        ]:
            if attr in self._node:
                setattr(self, f"_{attr}", self._node[attr])

        # RW attrs
        for attr in ["fixedCloudlets", "flexibleCloudlets"]:
            setattr(self, attr, self._node[attr])

        if self.nodeType == self.NodeType.DOCKER:
            try:
                self.docker_image = self._node["customitem"]["dockerName"]
            except KeyError:
                self.docker_image = ""

        # Copy our attributes as it came from API
        self.copy_self_as_from_api()

    def __init__(
        self,
        *,
        nodeType: NodeType = None,
        fixedCloudlets: int = 1,
        flexibleCloudlets: int = 2,
        node_group: "JelasticNodeGroup" = None,
        node_from_env: Dict[str, Any] = None,
    ) -> None:
        """
        Construct a JelasticNode from the outer data, or from
        """
        super().__init__()
        # By default, do not allow flexibleCloudlets' reduction
        self.allowFlexibleCloudletsReduction = False
        # Set initials
        self.fixedCloudlets = fixedCloudlets
        self.flexibleCloudlets = flexibleCloudlets

        if nodeType:
            self._nodeType = nodeType
            self._nodemission = nodeType.value

        if node_group:
            deprecation(
                "Node.__init__(): Passing node_group is deprecated; use attach_to_node_group instead",
            )
            self.attach_to_node_group(node_group)
            assert self.nodeGroup == node_group

        if node_from_env:
            deprecation(
                "Node.__init__(): Passing node_from_env is deprecated; use update_from_env_dict instead",
            )
            self.update_from_env_dict(node_from_env=node_from_env)
            assert self.is_from_api

    @property
    def links(self) -> List[Dict[str, Any]]:
        """
        Expose the inwards docker links, as they came from the API, for consumption from nodeGroup
        """
        try:
            return [
                n for n in self._node["customitem"]["dockerLinks"] if n["type"] == "IN"
            ]
        except (KeyError, AttributeError):
            return []

    def __str__(self) -> str:
        return f"JelasticNode id:{self.id}"

    def _set_cloudlets(self):
        """
        Set the cloudlets' count on that node
        """
        self.raise_unless_can_update_to_api()

        if not self._from_api or (
            self._from_api["fixedCloudlets"] != self.fixedCloudlets
            or self._from_api["flexibleCloudlets"] != self.flexibleCloudlets
        ):
            if self.flexibleCloudlets < self._from_api["flexibleCloudlets"]:
                if not self.allowFlexibleCloudletsReduction:
                    raise JelasticObjectException(
                        "flexibleCloudlets cannot be reduced without setting allowFlexibleCloudletsReduction to True before save()"
                    )
                else:
                    # Reset the authorization to False
                    self.allowFlexibleCloudletsReduction = False
            self.api._(
                "Environment.Control.SetCloudletsCountById",
                envName=self.envName,
                count=1,  # TODO: this is the nunber of _nodes
                nodeid=self.id,
                fixedCloudlets=self.fixedCloudlets,
                flexibleCloudlets=self.flexibleCloudlets,
            )
            self._from_api["fixedCloudlets"] = self.fixedCloudlets
            self._from_api["flexibleCloudlets"] = self.flexibleCloudlets

    @property
    def envVars(self):
        """
        Access the lazy-loaded vars from the nodeGroup, always
        """
        try:
            return self._nodeGroup.envVars
        except AttributeError:
            raise JelasticObjectException(
                "Cannot fetch envVars from node without nodeGroup (not from API ?)"
            )

    def raise_unless_can_update_to_api(self):
        """
        Check if we can update to API, or raise
        """
        if not hasattr(self, "envName"):
            raise JelasticObjectException(
                "Cannot update to API, use attach_to_node_group() before saving!"
            )

    def save_to_jelastic(self):
        """
        Mandatory _JelasticObject method, to save status to Jelastic
        """
        self._set_cloudlets()

    # Jelastic-related utilities
    def execute_commands(self, commands: List[str]) -> List[Dict[str, str]]:
        """
        Execute a list of commands in this node
        """
        if not isinstance(commands, list):
            raise TypeError("execute_commands() takes a a list of commands")

        self.raise_unless_can_update_to_api()

        command_list = [{"command": cmd, "params": ""} for cmd in commands]
        command_results = self.api._(
            "Environment.Control.ExecCmdById",
            envName=self.envName,
            nodeid=self.id,
            commandList=json.dumps(command_list),
        )
        return [
            {k: cr[k] for k in ["out", "result", "errOut"]}
            for cr in command_results["responses"]
        ]

    def execute_command(self, command: str) -> Dict[str, str]:
        """
        Execute a single command in this node
        """
        if not isinstance(command, str):
            raise TypeError("execute_command() takes a string as command")
        return self.execute_commands([command])[0]

    def read_file(self, path: str) -> str:
        """
        Read a file in a node
        """
        if not path:
            raise TypeError(f"path {path} cannot be empty")

        from .environment import JelasticEnvironment

        if self.nodeGroup._parent.status != JelasticEnvironment.Status.RUNNING:
            raise JelasticObjectException(
                "Files cannot be read from environments not running."
            )

        self.raise_unless_can_update_to_api()

        response = self.api._(
            "Environment.File.Read",
            envName=self.envName,
            nodeid=self.id,
            path=path,
        )
        return response["body"]



from ..exceptions import deprecation
from .jelasticobject import _JelasticAttribute as _JelAttr
from .jelasticobject import _JelasticObject, _JelAttrBool, _JelAttrStr
from .nodegroup import JelasticNodeGroup


class _JelasticVolume(_JelasticObject):
    """
    Represents a Jelastic Volume (CountainerVolume or MountPoint)
    """

    # 'parent'
    nodeGroup = _JelAttr(read_only=True, checked_for_differences=False)

    # Whether it was from the API or not
    is_new = _JelAttrBool()

    # Source nodeGroup / Where it is mounted _to_
    name = _JelAttrStr(read_only=True)
    path = _JelAttrStr(read_only=True)

    def __init__(
        self,
        *,
        node_group: JelasticNodeGroup = None,
        name: str = None,
        path: str = None,
    ) -> None:
        """
        Construct a _JelasticVolume from basic stuff
        """
        super().__init__()

        if node_group:
            deprecation(
                "_Volume.__init__(): Passing node_group is deprecated; use attach_to_node_group instead",
            )
            self.attach_to_node_group(node_group)

        self._name = name
        self._path = path
        self.is_new = True
        assert self.differs_from_api

    def attach_to_node_group(self, node_group: JelasticNodeGroup) -> None:
        """
        Connect node_group to mountPoint, and inversely
        """

        node_group.append_mount_point(self)

        self._envName = self._nodeGroup._parent.envName

    def save_to_jelastic(self):
        """
        Mandatory _JelasticObject method, to save status to Jelastic
        """
        # TODO

    def __str__(self):
        """
        String representation
        """
        return f"_JelasticVolume {self.path} on {self.nodeGroup}"



import jelapi
from jelapi.connector import JelasticAPIConnector


def test_jelapi_api_connector_is_instance():
    assert isinstance(jelapi.api_connector(), JelasticAPIConnector)


def test_jelapi_api_connector_has_an_httpx_client():
    assert jelapi.api_connector().client


def test_jelapi_api_connector_sets_api_url():
    jelapi.api_url = "https://api.example.org/2.0/"
    assert jelapi.api_connector().apiurl == "https://api.example.org/2.0/"


def test_jelapi_api_connector_got_token():
    jelapi.api_token = "new-secret"
    assert jelapi.api_connector().apidata["session"] == "new-secret"



import pytest
import respx
from httpx import Response, codes

from jelapi import JelasticAPIException
from jelapi.connector import JelasticAPIConnector

APIURL = "https://api.example.org/"


@respx.mock
def test_connector_successfull_empty_answer():
    """
    Check that a simple call with a 'result': 0 is successful
    """
    japic = JelasticAPIConnector(apiurl=APIURL, apitoken="string")

    getenvs_route = respx.post(f"{APIURL}environment/control/rest/getenvs").mock(
        return_value=Response(status_code=codes.OK, json={"result": 0})
    )
    japic._("Environment.Control.GetEnvs")
    assert getenvs_route.called


@respx.mock
def test_connector_failed_empty_answer():
    """
    Check that a simple call with a 'result': 1 raises
    """
    japic = JelasticAPIConnector(apiurl=APIURL, apitoken="string")

    getenvs_route = respx.post(f"{APIURL}environment/control/rest/getenvs").mock(
        return_value=Response(status_code=codes.OK, json={"result": 1})
    )
    with pytest.raises(JelasticAPIException):
        japic._("Environment.Control.GetEnvs")
        assert getenvs_route.called


@respx.mock
def test_connector_fails_on_non_OK_status():
    """
    Check that a simple call with a 'result': 0 but an HTTP status != 200 raises
    """
    japic = JelasticAPIConnector(apiurl=APIURL, apitoken="string")

    getenvs_route = respx.post(f"{APIURL}environment/control/rest/getenvs").mock(
        return_value=Response(status_code=codes.IM_A_TEAPOT, json={"result": 0})
    )
    with pytest.raises(JelasticAPIException):
        japic._("Environment.Control.GetEnvs")
        assert getenvs_route.called


def test_connector_refuses_wrongly_formatted_functions():
    """
    Check that functions need to be in the First.Second.Word format
    """
    japic = JelasticAPIConnector(apiurl=APIURL, apitoken="string")

    with pytest.raises(JelasticAPIException):
        # 3 dots
        japic._("Not.A.Function.Call")

    with pytest.raises(JelasticAPIException):
        # 2 dots
        japic._("Not_A.Function_Call")

    with respx.mock() as respx_mock:
        # 3 dots is a valid function
        # TODO: Perhaps we allow-list the functions we know
        uncontrolled_route = respx_mock.post(f"{APIURL}a/function/rest/call").mock(
            return_value=Response(status_code=200, json={"result": 0})
        )
        japic._("A.Function.Call")
        assert uncontrolled_route.called


def test_connector_is_not_functional_if_apidata_not_a_list():
    japic = JelasticAPIConnector(apiurl=None, apitoken=None)
    japic.apidata = "some-str"
    assert not japic.is_functional()


def test_connector_is_not_functional_if_apiurl_is_deleted():
    japic = JelasticAPIConnector(apiurl=None, apitoken=None)
    del japic.apiurl
    assert not japic.is_functional()


def test_connector_is_not_functional_without_api_url():
    japic = JelasticAPIConnector(apiurl="", apitoken="secret")
    assert not japic.is_functional()


def test_connector_is_not_functional_without_api_token():
    japic = JelasticAPIConnector(apiurl=APIURL, apitoken="")
    assert not japic.is_functional()


def test_connector_is_functional_with_both():
    japic = JelasticAPIConnector(apiurl=APIURL, apitoken="secret")
    assert japic.is_functional()



import factory
from faker_enum import EnumProvider

from tests.utils import get_standard_env, get_standard_node, get_standard_node_group

from . import classes

factory.Faker.add_provider(provider=EnumProvider)


class _JelasticEnvironmentFactory(factory.Factory):
    class Meta:
        model = classes.JelasticEnvironment


class _JelasticNodeGroupFactory(factory.Factory):
    class Meta:
        model = classes.JelasticNodeGroup

    nodeGroupType = factory.Faker(
        "enum", enum_cls=classes.JelasticNodeGroup.NodeGroupType
    )


class _JelasticNodeFactory(factory.Factory):
    class Meta:
        model = classes.JelasticNode

    nodeType = factory.Faker("enum", enum_cls=classes.JelasticNode.NodeType)


class _JelasticMountPointFactory(factory.Factory):
    class Meta:
        model = classes.JelasticMountPoint

    nodeType = factory.Faker("enum", enum_cls=classes.JelasticNode.NodeType)


class JelasticEnvironmentFactory(_JelasticEnvironmentFactory):
    @classmethod
    def _after_postgeneration(obj, instance, create, results=None):
        """
        Generate a standard Env
        """
        instance.update_from_env_dict(get_standard_env())
        ngs = {}
        for key in ["cp", "sqldb", "storage"]:
            ng = JelasticNodeGroupFactory()
            ng._nodeGroupType = next(
                (
                    ng
                    for ng in classes.JelasticNodeGroup.NodeGroupType
                    if ng.value == key
                ),
            )
            ng.attach_to_environment(instance)

            for n in ng.nodes:
                if key in ["cp", "sqldb"]:
                    n._nodeType = classes.JelasticNode.NodeType.DOCKER
                elif key == "storage":
                    n._nodeType = classes.JelasticNode.NodeType.STORAGE

                # Set different Ids
                if key == "cp":
                    n._id = 110
                if key == "sqldb":
                    n._id = 120
                if key == "storage":
                    n._id = 130

            ngs[key] = ng

        instance.nodeGroups = ngs
        instance.copy_self_as_from_api("nodeGroups")


class JelasticNodeFactory(_JelasticNodeFactory):
    @classmethod
    def _after_postgeneration(obj, instance, create, results=None):
        """
        Generate a standard Node
        """
        instance.update_from_env_dict(get_standard_node())
        assert instance.is_from_api


class JelasticNodeGroupFactory(_JelasticNodeGroupFactory):
    @classmethod
    def _after_postgeneration(obj, instance, create, results=None):
        """
        Generate a standard NodeGroup, with one node of a random nodeType
        It can be set according to needs afterwards.
        """
        instance.update_from_env_dict(
            get_standard_node_group(node_group_type=instance.nodeGroupType)
        )
        node = JelasticNodeFactory()

        # Set different Ids
        if instance.nodeGroupType.value == "cp":
            node._id = 110
        if instance.nodeGroupType.value == "sqldb":
            node._id = 120
        if instance.nodeGroupType.value == "storage":
            node._id = 130

        node.attach_to_node_group(instance)
        instance.copy_self_as_from_api("nodes")



from warnings import warn


class JelapiException(Exception):
    """
    Generic Jelapi Exception
    """

    pass


class JelasticAPIException(JelapiException):
    """
    Low-level API Exception
    """

    pass


class JelasticObjectException(JelapiException):
    """
    Object instanciation issue
    """

    pass


def deprecation(message: str) -> None:
    """
    Announce deprecated argumets, functions, classes
    """
    warn(message, DeprecationWarning, stacklevel=2)



import warnings
from copy import deepcopy
from unittest.mock import Mock

import pytest

from jelapi import api_connector as jelapic
from jelapi.classes import JelasticEnvironment, JelasticMountPoint, JelasticNodeGroup
from jelapi.exceptions import JelasticObjectException
from jelapi.factories import JelasticEnvironmentFactory, JelasticNodeGroupFactory

from .utils import get_standard_mount_point, get_standard_node, get_standard_node_group

jelenv = JelasticEnvironmentFactory()


def test_JelasticNodeGroup_simple_load():
    """
    JelasticNodeGroup can be instantiated as-is
    """
    JelasticNodeGroup()


def test_JelasticNodeGroup_no_api_call_unless_setup():
    """
    JelasticNodeGroup can be instantiated as-is
    """
    ng = JelasticNodeGroup()
    with pytest.raises(JelasticObjectException):
        ng.raise_unless_can_call_api()


def test_JelasticNodeGroup_with_enough_data():
    """
    JelasticNodeGroup can be instantiated, deprecated format
    """
    with warnings.catch_warnings(record=True) as warns:
        # From API
        j1 = JelasticNodeGroup(
            nodeGroupType=JelasticNodeGroup.NodeGroupType.APPLICATION_SERVER,
            parent=jelenv,
            node_group_from_env=get_standard_node_group(),
        )
        assert j1.is_from_api
        assert len(warns) == 2

    with warnings.catch_warnings(record=True) as warns:
        j2 = JelasticNodeGroup(
            parent=jelenv,
            nodeGroupType=JelasticNodeGroup.NodeGroupType.SQL_DATABASE,
        )
        assert not j2.is_from_api
        assert len(warns) == 1


def test_JelasticNodeGroup_with_missing_data():
    """
    JelasticNodeGroup cannot be instantiated with missing attributes
    """
    for musthavekey in ["name"]:
        nodegroup = get_standard_node_group()
        del nodegroup[musthavekey]
        ng = JelasticNodeGroup()
        with pytest.raises(KeyError):
            # missing name (alone) dies
            ng.update_from_env_dict(nodegroup)


def test_JelasticNodeGroup_factory():
    """
    Factory works
    """
    node_group = JelasticNodeGroupFactory()
    assert len(node_group.nodes) > 0
    assert node_group.is_from_api


def test_JelasticNodeGroup_immutable_data():
    """
    Doesn't differ from API at build
    """
    node_group = JelasticNodeGroupFactory(
        nodeGroupType=JelasticNodeGroup.NodeGroupType.APPLICATION_SERVER
    )
    assert str(node_group) == "JelasticNodeGroup cp"

    with pytest.raises(AttributeError):
        #  nodeGroup cannot be changed to string
        node_group.nodeGroupType = "sqldb"
    with pytest.raises(AttributeError):
        #  nodeGroup cannot be changed to enum either
        node_group.nodeGroupType = JelasticNodeGroup.NodeGroupType.SQL_DATABASE


def test_JelasticNodeGroup_envVars_refreshes_from_API():
    """
    Getting the envVars gets us an API call
    """
    node_group = JelasticNodeGroupFactory()
    assert node_group.nodeGroupType
    node_group.attach_to_environment(JelasticEnvironmentFactory())
    node_group.raise_unless_can_call_api()

    jelapic()._ = Mock(
        return_value={"object": {"VAR": "value"}},
    )
    assert "VAR" in node_group.envVars
    assert node_group.envVars["VAR"] == "value"
    # As we lazy-load, accessing the dict will call the API once
    jelapic()._.assert_called_once()
    assert not node_group.differs_from_api()


def test_JelasticNodeGroup_envVars_raises_if_set_without_fetch():
    """
    Saving a faked envVars without fetch will raise
    """
    node_group = JelasticNodeGroupFactory()
    node_group.attach_to_environment(jelenv)
    # Make sure they never were fetched
    node_group._envVars_need_fetching = True
    node_group._envVars = {"ID": "evil"}

    assert node_group._envVars != node_group._from_api["_envVars"]

    with pytest.raises(JelasticObjectException):
        node_group.save()


def test_JelasticNodeGroup_envVars_raises_if_set_empty():
    """
    Saving a faked envVars without fetch will raise
    """
    node_group = JelasticNodeGroupFactory()
    node_group.attach_to_environment(JelasticEnvironmentFactory())
    jelapic()._ = Mock(
        return_value={"object": {"VAR": "value"}},
    )
    # fetch it
    node_group.envVars
    # empty it
    node_group._envVars = {}
    with pytest.raises(JelasticObjectException):
        node_group.save()


def test_JelasticNodeGroup_envVars_raises_if_env_is_not_running():
    """
    Fetching envVars raises on envs not runnirg
    """
    from jelapi.classes import JelasticEnvironment

    JelStatus = JelasticEnvironment.Status
    for status in JelStatus:
        jelenv_local = deepcopy(jelenv)
        jelenv_local.status = status
        node_group = JelasticNodeGroupFactory()
        node_group.attach_to_environment(jelenv_local)
        jelapic()._ = Mock(
            return_value={"object": {"VAR": "value"}},
        )
        if status in [
            JelStatus.RUNNING,
            JelStatus.CREATING,
            JelStatus.CLONING,
        ]:
            node_group.envVars
        else:
            with pytest.raises(JelasticObjectException):
                node_group.envVars


def test_JelasticNodeGroup_envVars_updates():
    node_group = JelasticNodeGroupFactory()
    node_group.attach_to_environment(JelasticEnvironmentFactory())

    jelapic()._ = Mock(
        return_value={"object": {"VAR": "value"}},
    )
    assert node_group.envVars["VAR"] == "value"
    # As we lazy-load, accessing the dict will call the API once
    jelapic()._.assert_called_once()

    node_group.envVars["NEWVAR"] = "revolution"
    assert node_group.differs_from_api()
    jelapic()._ = Mock(return_value={"result": 0})
    # Save the addition, it will work, and call one set
    node_group.save()
    assert not node_group.differs_from_api()
    jelapic()._.assert_called_once()

    jelapic()._.reset_mock()
    del node_group.envVars["NEWVAR"]
    # Save the removal, it will work, and call one set
    node_group.save()
    jelapic()._.assert_called_once()
    assert not node_group.differs_from_api()

    # Now test both addition and removal, this will do a single "set"
    node_group.envVars["NEWVAR"] = "new"
    del node_group.envVars["VAR"]
    jelapic()._.reset_mock()
    node_group.save()
    jelapic()._.assert_called_once()
    assert not node_group.differs_from_api()


def test_JelasticNodeGroup_displayName_update():
    """
    Update displayName
    """
    node_group = JelasticNodeGroupFactory()
    node_group.attach_to_environment(JelasticEnvironmentFactory())

    assert not node_group.differs_from_api()
    node_group.displayName = "A new name for a new day"
    assert node_group.differs_from_api()

    jelapic()._ = Mock()
    node_group.save()
    assert not node_group.differs_from_api()
    jelapic()._.assert_called_once()


def test_JelasticNodeGroup_SLB_update():
    """
    Update SLB status
    """
    node_group = JelasticNodeGroupFactory()
    node_group.attach_to_environment(JelasticEnvironmentFactory())

    assert not node_group.differs_from_api()
    # Toggle the status
    node_group.isSLBAccessEnabled = not node_group.isSLBAccessEnabled
    assert node_group.differs_from_api()

    jelapic()._ = Mock()
    node_group.save()
    assert not node_group.differs_from_api()
    jelapic()._.assert_called_once()


def test_JelasticNodeGroup_redeploy():
    """
    NodeGroups can be redeployed
    """
    node_group = JelasticNodeGroupFactory()
    node_group.attach_to_environment(JelasticEnvironmentFactory())

    assert not node_group.differs_from_api()
    node_group.redeploy(docker_tag="latest")


def test_JelasticNodeGroup_read_file():
    """
    We can gather a single file in a nodegroup
    """
    node_group = JelasticNodeGroupFactory()
    node_group.attach_to_environment(JelasticEnvironmentFactory())

    jelapic()._ = Mock(
        return_value={
            "body": "Text content",
            "result": 0,
        },
    )
    with pytest.raises(TypeError):
        # We can't read no file
        node_group.read_file("")

    for status in JelasticEnvironment.Status:
        node_group._parent.status = status

        jelapic()._.reset_mock()
        if status != JelasticEnvironment.Status.RUNNING:
            with pytest.raises(JelasticObjectException):
                node_group.read_file("/tmp/test")
        else:
            body = node_group.read_file("/tmp/test")
            jelapic()._.assert_called_once()
            assert body == "Text content"


def test_JelasticNodeGroup_get_mountPoints():
    """
    We can get the list of mountPoints
    """
    jelenv = JelasticEnvironmentFactory()
    node_group = jelenv.nodeGroups["cp"]

    assert node_group._mountPoints_need_fetching

    jelapic()._ = Mock(
        return_value={
            "array": [
                get_standard_mount_point(
                    source_node_id=jelenv.nodeGroups["storage"].nodes[0].id
                )
            ]
        },
    )
    assert len(node_group.mountPoints) == 1
    jelapic()._.assert_called_once()
    # Now assume it changed on the API
    jelapic()._ = Mock(
        return_value={"array": []},
    )
    # It did not change, and the API was not called
    assert node_group.mountPoints != []
    jelapic()._.assert_not_called()

    # There's no way to force-refresh, currently:
    with pytest.raises(TypeError):
        node_group._mountPoints = None
    node_group._mountPoints = []
    # It's still 0
    assert len(node_group.mountPoints) == 0
    jelapic()._.assert_not_called()


def test_JelasticNodeGroup_add_remove_mountPoints():
    """
    We can add mountPoints
    """
    # Instantiate a somewhat realistic environment
    cp_node_group = JelasticNodeGroupFactory(
        nodeGroupType=JelasticNodeGroup.NodeGroupType.APPLICATION_SERVER
    )
    storage_node_group = JelasticNodeGroupFactory(
        nodeGroupType=JelasticNodeGroup.NodeGroupType.SQL_DATABASE
    )

    jelenv = JelasticEnvironmentFactory()
    cp_node_group.attach_to_environment(jelenv)
    storage_node_group.attach_to_environment(jelenv)

    jelapic()._ = Mock(
        return_value={"array": []},
    )
    assert cp_node_group.mountPoints == []
    assert not cp_node_group._mountPoints_need_fetching

    # As we fetched, to append
    jelapic()._.assert_called_once()

    jmp = JelasticMountPoint(
        name="test",
        path="/tmp/test",
        sourcePath="/srv",
        sourceNode=storage_node_group.nodes[0],
    )
    jmp.attach_to_node_group(cp_node_group)
    assert not jmp.is_from_api

    assert len(cp_node_group.mountPoints) == 1

    jelapic()._.reset_mock()
    cp_node_group.save()
    jelapic()._.assert_called_once()
    # It made it "from API"
    assert jmp.is_from_api
    assert len(cp_node_group._from_api["_mountPoints"]) == 1
    assert cp_node_group._from_api["_mountPoints"][0].is_from_api

    # Add another one, with the same path
    jmp2 = JelasticMountPoint(
        name="test2",
        path="/tmp/test",
        sourcePath="/srv/test2",
        sourceNode=storage_node_group.nodes[0],
    )
    jmp2.attach_to_node_group(cp_node_group)
    assert not jmp2.is_from_api

    # It will override the previous one, so…
    assert len(cp_node_group.mountPoints) == 1

    # Saving shall work
    jelapic()._.reset_mock()
    cp_node_group.save()
    jelapic()._.assert_called_once()

    # Remove the only one
    del cp_node_group.mountPoints[0]
    assert len(cp_node_group.mountPoints) == 0
    assert len(cp_node_group._from_api["_mountPoints"]) == 1
    assert cp_node_group._from_api["_mountPoints"][0].is_from_api

    # So the save will also remove the first one
    jelapic()._.reset_mock()
    cp_node_group.save()
    jelapic()._.assert_called_once()


def test_JelasticNodeGroup_links():
    """
    Getting the links works without further ado
    """
    ng = JelasticNodeGroupFactory()
    assert len(ng.links) == 0


def test_JelasticNodeGroup_links_cannot_be_fetched_without_nodes():
    """
    Getting the links doesn't work without nodes
    """
    ng = JelasticNodeGroupFactory()
    ng.nodes = []
    with pytest.raises(JelasticObjectException):
        ng.links


def test_JelasticNodeGroup_adding_links_means_topology_change():
    """
    Getting the links doesn't work without nodes
    """
    cpng = JelasticNodeGroupFactory(
        nodeGroupType=JelasticNodeGroup.NodeGroupType.APPLICATION_SERVER
    )
    sqldbng = JelasticNodeGroupFactory(
        nodeGroupType=JelasticNodeGroup.NodeGroupType.SQL_DATABASE
    )
    cpng.attach_to_environment(jelenv)
    sqldbng.attach_to_environment(jelenv)
    assert len(cpng.links) == 0

    cpng.links["SQLDB"] = JelasticNodeGroup.NodeGroupType.SQL_DATABASE
    assert len(cpng.links) == 1
    assert cpng.needs_topology_update()


def test_JelasticNodeGroup_links_non_empty():
    """
    Getting the links works with a node that has a link
    """
    jelenv = JelasticEnvironmentFactory()
    cpng = jelenv.nodeGroups["cp"]
    sqldbng = jelenv.nodeGroups["sqldb"]

    ndict = get_standard_node()
    ndict["id"] = cpng.nodes[0].id

    ndict["customitem"] = {
        "dockerLinks": [
            {"type": "IN", "sourceNodeId": sqldbng.nodes[0].id, "alias": "SQLDB"},
            {"type": "OUT", "sourceNodeId": 154},
        ]
    }
    cpng.nodes[0].update_from_env_dict(ndict)

    assert len(cpng.links) == 1
    assert "SQLDB" in cpng.links
    assert cpng.links["SQLDB"] == JelasticNodeGroup.NodeGroupType.SQL_DATABASE


def test_JelasticNodeGroup_containerVolumes():
    """
    Get container volumes
    """
    ng = JelasticNodeGroupFactory()
    ng.attach_to_environment(jelenv)

    ng._mountPoints_need_fetching = False
    ng._mountPoints = []

    jelapic()._ = Mock(
        return_value={"object": ["/tmp/volume1", "/tmp/volume2"]},
    )
    assert len(ng.containerVolumes) == 2
    assert jelapic()._.called_once()


def test_JelasticNodeGroup_containerVolumes_add_remove():
    """
    Get container volumes
    """
    ng = JelasticNodeGroupFactory()
    ng.attach_to_environment(jelenv)

    ng._mountPoints_need_fetching = False
    ng._mountPoints = []

    ng.copy_self_as_from_api("_mountPoints")
    ng._containerVolumes = ["/tmp/volume1", "/tmp/volume2"]
    ng.copy_self_as_from_api("_containerVolumes")

    ng.containerVolumes.append("/srv")
    assert ng.differs_from_api()

    jelapic()._ = Mock()
    ng.save()
    # There was one add
    jelapic()._.assert_called_once()
    assert not ng.differs_from_api()
    assert len(ng.containerVolumes) == 3

    del ng.containerVolumes[2]
    assert len(ng.containerVolumes) == 2

    jelapic()._.reset_mock()
    ng.save()
    # There was one removal
    jelapic()._.assert_called_once()
    assert not ng.differs_from_api()
    assert len(ng.containerVolumes) == len(ng._containerVolumes) == 2

    #  Add an identical
    ng.containerVolumes.append(ng.containerVolumes[0])
    with pytest.raises(JelasticObjectException):
        ng.save()


def test_JelasticNodeGroup_topology_without_nodes():
    """
    These fail, we need the node for lots of information
    """
    ng = JelasticNodeGroupFactory()
    ng.nodes = []
    with pytest.raises(JelasticObjectException):
        ng.get_topology()


def test_JelasticNodeGroup_topology():
    """
    Get a full'er topology
    """
    jelenv = JelasticEnvironmentFactory()
    cp_node_group = jelenv.nodeGroups["cp"]

    cp_node_group.links["SQLDB"] = JelasticNodeGroup.NodeGroupType.SQL_DATABASE
    cp_node_group.nodes[0].docker_registry = {"url": "https://docker.example.com/"}
    cp_node_group.nodes[0].docker_image = "example_image:tag"
    assert len(cp_node_group.nodes) > 0

    cp_node_group._envVars_need_fetching = False
    cp_node_group.envVars["HOSTNAME"] = "https://example.com"

    # Now get it
    topology = cp_node_group.get_topology()
    assert topology["count"] == len(cp_node_group.nodes)
    # Check that the links' syntax got computed correctly
    assert topology["links"] == ["sqldb:SQLDB"]

    cp_node_group.links["BROKEN"] = "sqldb"
    with pytest.raises(TypeError):
        cp_node_group.get_topology()


def test_JelasticNodeGroup_duplicate_mountPoints():
    """
    Test we can't save duplicated mount points
    """
    j = JelasticEnvironmentFactory()
    cpng = j.nodeGroups["cp"]

    # Assume they got fetched
    cpng._mountPoints_need_fetching = False
    # Attach one.
    mount = JelasticMountPoint(
        name="Test",
        path="/srv",
        sourceNode=j.nodeGroups["storage"].nodes[0],
        sourcePath="/srv/src",
    )
    mount.attach_to_node_group(cpng)
    assert len(cpng.mountPoints) == 1

    # Now pretend to add another one, but cheat
    cpng.mountPoints.append(mount)

    with pytest.raises(JelasticObjectException):
        cpng._save_mount_points()


def test_JelasticNodeGroup_copy_from_api():
    """
    Can we copy a nodeGroup
    """
    j = JelasticEnvironmentFactory()
    cpng = j.nodeGroups["cp"]

    # Let's assume they were fetched
    cpng._mountPoints_need_fetching = False
    cpng._containerVolumes_need_fetching = False
    cpng._envVars_need_fetching = False

    assert cpng.is_from_api
    cpng2 = cpng.archive_from_api()
    assert not cpng2.is_from_api



import json
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List

from ..exceptions import JelasticObjectException, deprecation
from .jelasticobject import _JelasticAttribute as _JelAttr
from .jelasticobject import (
    _JelasticObject,
    _JelAttrBool,
    _JelAttrDict,
    _JelAttrList,
    _JelAttrStr,
)

if TYPE_CHECKING:  # pragma: no cover
    from .environment import JelasticEnvironment
    from .mountpoint import JelasticMountPoint
    from .node import JelasticNode


class JelasticNodeGroup(_JelasticObject):
    """
    Represents a Jelastic NodeGroup, a sort of collection of Nodes within an environment
    """

    class NodeGroupType(Enum):
        """
        Standard NodeGroups
        """

        LOAD_BALANCER = "bl"
        APPLICATION_SERVER = "cp"
        CACHE = "cache"
        SQL_DATABASE = "sqldb"
        NOSQL_DATABASE = "nosqldb"
        STORAGE_CONTAINER = "storage"

    nodeGroupType = _JelAttr(read_only=True)
    envName = _JelAttrStr(read_only=True)

    isSLBAccessEnabled = _JelAttrBool()
    displayName = _JelAttrStr()
    nodes = _JelAttrList(checked_for_differences=False)

    # Variables
    _envVars = (
        _JelAttrDict()
    )  # this is the JelAttr, use envVars to access them through lazy loading
    _envVars_need_fetching = _JelAttrBool(checked_for_differences=False)

    _mountPoints = _JelAttrList(
        checked_for_differences=False
    )  # this is the JelAttr, use mountPoints to access them through lazy loading
    # Boolean flag to help lazy-load
    _mountPoints_need_fetching = _JelAttrBool(checked_for_differences=False)

    _containerVolumes = (
        _JelAttrList()
    )  # this is the JelAttr, use containerVolumes to access them through lazy loading
    # Boolean flag to help lazy-load
    _containerVolumes_need_fetching = _JelAttrBool(checked_for_differences=False)

    _links = (
        _JelAttrDict()
    )  # this is the JelAttr, use links to access them through lazy loading

    @property
    def envVars(self):
        """
        Lazy load envVars when they're accessed
        """
        from .environment import JelasticEnvironment

        JelStatus = JelasticEnvironment.Status

        if hasattr(self, "_parent") and self._parent.status not in [
            JelStatus.RUNNING,
            JelStatus.CREATING,
            JelStatus.CLONING,
        ]:
            raise JelasticObjectException(
                "envVars cannot be gathered on environments not running"
            )
        if self._envVars_need_fetching:
            self.raise_unless_can_call_api()

            response = self.api._(
                "Environment.Control.GetContainerEnvVarsByGroup",
                envName=self.envName,
                nodeGroup=self.nodeGroupType.value,
            )
            self._envVars = response["object"]
            self._envVars_need_fetching = False
            self.copy_self_as_from_api("_envVars")
        return self._envVars

    def _set_env_vars(self):
        """
        Set the modified envVars
        """
        # Only set them if they were fetched first
        if self._envVars_need_fetching:
            if self._envVars != {}:
                raise JelasticObjectException(
                    "envVars cannot be saved if not fetched first (no blind set)"
                )
        else:
            if len(self._envVars) == 0 and len(self._from_api["_envVars"]) > 0:
                raise JelasticObjectException(
                    "envVars cannot be set to empty (no wipe out)"
                )
            if self._from_api["_envVars"] != self._envVars:
                self.raise_unless_can_call_api()
                self.api._(
                    "Environment.Control.SetContainerEnvVarsByGroup",
                    envName=self.envName,
                    nodeGroup=self.nodeGroupType.value,
                    data=json.dumps(self._envVars),
                )
                self.copy_self_as_from_api("_envVars")

    @property
    def mountPoints(self) -> List["JelasticMountPoint"]:
        """
        Lazy load mountPoints when they're accessed
        """
        self.raise_unless_can_call_api()

        # from .environment import JelasticEnvironment
        from .mountpoint import JelasticMountPoint

        # JelStatus = JelasticEnvironment.Status
        #
        # if self._parent.status not in [
        #     JelStatus.RUNNING,
        #     JelStatus.CREATING,
        #     JelStatus.CLONING,
        # ]:
        #     raise JelasticObjectException(
        #         "envVars cannot be gathered on environments not running"
        #     )
        if self._mountPoints_need_fetching:
            response = self.api._(
                "Environment.File.GetMountPoints",
                envName=self.envName,
                nodeGroup=self.nodeGroupType.value,
            )

            for mpdict in response["array"]:
                mp = JelasticMountPoint()
                mp.attach_to_node_group(self)
                mp.update_from_env_dict(mpdict)

            self._mountPoints_need_fetching = False
            self.copy_self_as_from_api("_mountPoints")
        return self._mountPoints

    @property
    def links(self) -> Dict[str, NodeGroupType]:
        """
        "IN"wards links dict {"key": ng}
        """
        if len(self.nodes) == 0:
            raise JelasticObjectException("Links can't be fetched without nodes")
        if not hasattr(self, "_links"):
            self._links = {}
            for link in self.nodes[0].links:
                for node_group in self._parent.nodeGroups.values():
                    if link["sourceNodeId"] in [n.id for n in node_group.nodes]:
                        self._links[link["alias"]] = node_group.nodeGroupType
                        break
            self.copy_self_as_from_api("_links")

        return self._links

    def _save_mount_points(self):
        """
        Verify that the mount points have not changed, apply the changes
        """
        # Only check them if they were accessed
        if not self._mountPoints_need_fetching:
            mountpaths = [m.path for m in self.mountPoints]
            if len(set(mountpaths)) != len(mountpaths):
                raise JelasticObjectException(
                    f"Duplicate MountPoints won't work {','.join(mountpaths)}"
                )
            # Delete the obsolete mountpaths
            for mp in self._from_api.get("_mountPoints", []):
                if mp.path not in mountpaths:
                    mp.del_from_api()
                mp.save()

            # Create the new mountpaths
            for mp in self.mountPoints:
                if not mp.is_from_api:
                    mp.add_to_api()
                mp.save()
                assert not mp.differs_from_api()

            for mp in self.mountPoints:
                mp.copy_self_as_from_api()

        self.copy_self_as_from_api("_mountPoints")
        for mp in self._from_api.get("_mountPoints", []):
            mp.copy_self_as_from_api()

    @property
    def containerVolumes(self) -> List[str]:
        """
        Lazy load containerVolumes when they're accessed
        """
        self.raise_unless_can_call_api()

        if self._containerVolumes_need_fetching:
            response = self.api._(
                "Environment.Control.GetContainerVolumesByGroup",
                envName=self.envName,
                nodeGroup=self.nodeGroupType.value,
            )
            # We need to exclude the mountPoints
            self._containerVolumes = [
                cv
                for cv in response["object"]
                if cv not in [mp.path for mp in self.mountPoints]
            ]
            self._containerVolumes_need_fetching = False
            self.copy_self_as_from_api("_containerVolumes")
        return self._containerVolumes

    def _save_container_volumes(self):
        """
        Verify that the container volumes have not changed, apply the changes
        """
        self.raise_unless_can_call_api()

        # Only check them if they were accessed
        if not self._containerVolumes_need_fetching:
            if len(set(self.containerVolumes)) != len(self.containerVolumes):
                raise JelasticObjectException(
                    f"Duplicate Container Volumes won't work {','.join(self.containerVolumes)}"
                )
            # Delete the obsolete containerVolumes
            toremove = [
                cv
                for cv in self._from_api.get("_containerVolumes", [])
                if cv not in self.containerVolumes
            ]
            if len(toremove) > 0:
                self.api._(
                    "Environment.Control.RemoveContainerVolumes",
                    envName=self.envName,
                    nodeGroup=self.nodeGroupType.value,
                    volumes=json.dumps(toremove),
                )

            # Create the new mountpaths
            toadd = [
                cv
                for cv in self.containerVolumes
                if cv not in self._from_api.get("_containerVolumes", [])
            ]
            if len(toadd) > 0:
                self.api._(
                    "Environment.Control.AddContainerVolumes",
                    envName=self.envName,
                    nodeGroup=self.nodeGroupType.value,
                    volumes=json.dumps(toadd),
                )
            self.copy_self_as_from_api("_containerVolumes")

    def get_topology(self) -> Dict[str, Any]:
        """
        Return the "NodeGroup / Nodes" topology, as consumed by ChangeTopology
        """
        if len(self.nodes) == 0:
            raise JelasticObjectException("Can't get topology for an empty nodeGroup")
        # return {"count": , "restartDelay": 0, "nod"}
        node0 = self.nodes[0]
        topology = {
            "count": len(self.nodes),
            "restartDelay": 0,
            "displayName": self.displayName,
            "isSLBAccessEnabled": self.isSLBAccessEnabled,
            "nodeGroup": self.nodeGroupType.value,
            "nodeType": node0.nodeType.value,
            "mission": node0.nodemission,
            "fixedCloudlets": node0.fixedCloudlets,
            "flexibleCloudlets": node0.flexibleCloudlets,
        }
        from .node import JelasticNode

        if node0.nodeType == JelasticNode.NodeType.DOCKER:
            if hasattr(node0, "docker_registry"):
                topology["registry"] = node0.docker_registry

            if hasattr(node0, "docker_image") and node0.docker_image:
                topology["image"] = node0.docker_image

        # The important links
        links = []
        for key, ngtype in self.links.items():
            if not isinstance(ngtype, self.NodeGroupType):
                raise TypeError(
                    f"Links' values must be of type NodeGroupType ({ngtype})"
                )
            # Find the correct node_group value ("storage" or "sqldb") in the parent's
            node_group = next(
                (
                    ng.nodeGroupType.value
                    for ng in self._parent.nodeGroups.values()
                    if ng.nodeGroupType == ngtype
                ),
            )
            links.append(f"{node_group}:{key}")

        if links:
            topology["links"] = links

        # The environment variables, if we got them.
        if not self._envVars_need_fetching:
            if self._envVars:
                topology["env"] = self._envVars
        return topology

    def needs_topology_update(self):
        """
        Whether the ng needs a topology update from the environment
        """
        if not hasattr(self, "_links"):
            # Never fetched, no need
            return False
        return self._from_api["_links"] != self._links

    def append_node(self, node: "JelasticNode") -> None:
        """
        Called from node, allow to append a node to our list
        """
        self.nodes.append(node)
        node._nodeGroup = self

    def append_mount_point(self, mount_point: "JelasticMountPoint") -> None:
        """
        Called from mount_point, allow to append a mountPoint to our list
        """
        # Assume we want to force this one
        for mp in self._mountPoints:
            if mp.path == mount_point.path:
                self._mountPoints.remove(mp)
        self._mountPoints.append(mount_point)
        mount_point._nodeGroup = self

    def attach_to_environment(self, environment: "JelasticEnvironment") -> None:
        """
        Set the parent environment, if nodeGroup is correctly setup
        """
        # Make sure the parent also has this nodeGroup listed
        if hasattr(self, "nodeGroupType"):
            self._parent = environment

            environment.attach_node_group(self)

            # Read-only attributes
            self._envName = self._parent.envName

    def update_from_env_dict(self, node_group_from_env: Dict[str, Any]) -> None:
        """
        Construct/Update our object from the structure
        """
        # Allow exploration of the returned object, but don't act on it.
        self._node_group = node_group_from_env

        self._nodeGroupType = next(
            (ng for ng in self.NodeGroupType if ng.value == self._node_group["name"]),
        )

        # R/W attributes
        self._displayName = self._node_group.get(
            "displayName", ""
        )  # Apparently optional
        self._isSLBAccessEnabled = self._node_group.get("isSLBAccessEnabled")

        # These two need fetching now (it was from API)
        self._mountPoints_need_fetching = True
        self._containerVolumes_need_fetching = True
        self._envVars_need_fetching = True

        # Copy our attributes as it came from API
        self.copy_self_as_from_api()

    def raise_unless_can_call_api(self):
        """
        Check if we can update to API, or raise
        """
        if not hasattr(self, "envName") or not hasattr(self, "nodeGroupType"):
            raise JelasticObjectException(
                "Cannot update to API, use attach_to_environment() before saving!"
            )

    def __init__(
        self,
        *,
        parent: "JelasticEnvironment" = None,
        node_group_from_env: Dict[str, Any] = None,
        nodeGroupType: NodeGroupType = None,
    ) -> None:
        """
        Construct a JelasticNodeGroup from the outer data
        """
        super().__init__()

        # Instantiate some empty, or default
        self._displayName = ""
        self._isSLBAccessEnabled = False

        self.nodes = []
        self._mountPoints = []
        self._mountPoints_need_fetching = False
        self._containerVolumes = []
        self._containerVolumes_need_fetching = False

        self._envVars = {}
        self._envVars_need_fetching = False

        if nodeGroupType:
            # Construct a node Group out of the blue
            self._nodeGroupType = nodeGroupType

            if parent:
                deprecation(
                    "NodeGroup.__init__(): Passing parent is deprecated; use attach_to_environment instead",
                )
                self.attach_to_environment(parent)
                assert self._parent == parent

        if node_group_from_env:
            deprecation(
                "NodeGroup.__init__(): Passing node_group_from_env is deprecated; use update_from_env_dict instead",
            )
            self.update_from_env_dict(node_group_from_env=node_group_from_env)
            assert self.is_from_api

    def before_archive_from_api(self):
        """
        When archiving from API, fetch mountPoints, containerVolumes and envVars
        """
        self.mountPoints
        self.containerVolumes
        self.envVars

    def _apply_data(self):
        """
        Use "ApplyData" to save all the data we _can_ save
        """
        self.raise_unless_can_call_api()

        #  Prepare data attr for ApplyData call
        data = {}
        for attr in ["displayName", "isSLBAccessEnabled"]:
            v = getattr(self, attr)
            if attr not in self._from_api or self._from_api[attr] != v:
                data[attr] = v
        if len(data) > 0:
            # Jelastic API 5.9
            self.api._(
                "Environment.Control.ApplyNodeGroupData",
                envName=self.envName,
                nodeGroup=self.nodeGroupType.value,
                data=json.dumps(data),
            )
            for k in data.keys():
                self._from_api[k] = data[k]

    def save_to_jelastic(self):
        """
        Mandatory _JelasticObject method, to save status to Jelastic
        """
        self._apply_data()
        self._set_env_vars()
        for n in self.nodes:
            n.save()
        self.copy_self_as_from_api("nodes")
        self._save_container_volumes()
        self._save_mount_points()

    def __str__(self) -> str:
        """
        String representation
        """
        return f"JelasticNodeGroup {self.nodeGroupType.value}"

    # Jelastic-related utilites
    def read_file(self, path: str) -> str:
        """
        Read a file in a nodeGroup
        """
        self.raise_unless_can_call_api()

        if not path:
            raise TypeError(f"path {path} cannot be empty")

        from .environment import JelasticEnvironment

        if self._parent.status != JelasticEnvironment.Status.RUNNING:
            raise JelasticObjectException(
                "Files cannot be read from environments not running."
            )
        response = self.api._(
            "Environment.File.Read",
            envName=self.envName,
            nodeGroup=self.nodeGroupType.value,
            path=path,
        )
        return response["body"]

    def redeploy(self, docker_tag: str = "latest"):
        """
        Redeploy a nodeGroup to a certain docker tag
        """
        self.raise_unless_can_call_api()

        self.api._(
            "Environment.Control.RedeployContainersByGroup",
            envName=self.envName,
            nodeGroup=self.nodeGroupType.value,
            tag=docker_tag,
        )



import warnings
from unittest.mock import Mock

import pytest

from jelapi import api_connector as jelapic
from jelapi.classes.mountpoint import JelasticMountPoint
from jelapi.exceptions import JelasticObjectException
from jelapi.factories import JelasticEnvironmentFactory

from .utils import get_standard_mount_point

# Create default environment
jelenv = JelasticEnvironmentFactory()
cp_node_group = jelenv.nodeGroups["cp"]
storage_node_group = jelenv.nodeGroups["storage"]


def test_JelasticMountPoint_simple():
    """
    Test we cannot instantiate in various situations
    """
    JelasticMountPoint()


def test_JelasticMountPoint_deprecations():
    """
    Test we cannot instantiate in various situations
    """

    with warnings.catch_warnings(record=True) as warns:
        JelasticMountPoint(node_group=cp_node_group)
        assert len(warns) == 1

    with warnings.catch_warnings(record=True) as warns:
        JelasticMountPoint(
            node_group=cp_node_group,
            mount_point_from_api=get_standard_mount_point(
                storage_node_group.nodes[0].id
            ),
        )
        assert len(warns) == 2

    # In "deprecated mode", both are needed
    with pytest.raises(TypeError):
        JelasticMountPoint(mount_point_from_api=get_standard_mount_point())


def test_JelasticMountPoint_init_from_api():
    """
    Test we can instantiate this, and take the str representation
    """
    # The Environment needs to have the target mount point ID as node
    jmp = JelasticMountPoint()
    with pytest.raises(JelasticObjectException):
        # Cannot update_from_env before attaching to node_group
        jmp.update_from_env_dict(
            get_standard_mount_point(source_node_id=storage_node_group.nodes[0].id)
        )
    jmp.attach_to_node_group(cp_node_group)
    jmp.update_from_env_dict(
        get_standard_mount_point(source_node_id=storage_node_group.nodes[0].id)
    )
    assert jmp.is_from_api
    assert str(jmp) != ""


def test_JelasticMountPoint_init_as_new():
    """
    Test we can instantiate this as new mountPoint
    """
    jmp = JelasticMountPoint(
        name="test name",
        path="/tmp/test1",
        sourceNode=storage_node_group.nodes[0],
        sourcePath="/tmp/destination",
    )
    jmp.attach_to_node_group(cp_node_group)
    assert not jmp.is_from_api


def test_JelasticMountPoint_init_no_source_avail():
    """
    Test we cannot instantiate if the api points to an unknown node
    """
    jmp = JelasticMountPoint()
    jmp.attach_to_node_group(cp_node_group)
    with pytest.raises(JelasticObjectException):
        jmp.update_from_env_dict(
            get_standard_mount_point(
                source_node_id=storage_node_group.nodes[0].id + 10
            ),
        )


def test_JelasticMountPoint_cannot_add_but_del_if_came_from_api():
    """
    We can't add a mount point that came from the API
    """

    jmp = JelasticMountPoint()
    jmp.attach_to_node_group(cp_node_group)
    jmp.update_from_env_dict(
        get_standard_mount_point(source_node_id=storage_node_group.nodes[0].id),
    )
    assert jmp.is_from_api
    with pytest.raises(JelasticObjectException):
        jmp.add_to_api()

    jelapic()._ = Mock()
    jmp.del_from_api()
    jelapic()._.assert_called_once()


def test_JelasticMountPoint_cannot_del_but_add_if_new():
    """
    We can't add a mount point that came from the API
    """
    jmp = JelasticMountPoint(
        name="test name",
        path="/tmp/test1",
        sourceNode=storage_node_group.nodes[0],
        sourcePath="/tmp/destination",
    )
    jmp.attach_to_node_group(cp_node_group)
    assert not jmp.is_from_api
    with pytest.raises(JelasticObjectException):
        jmp.del_from_api()

    jelapic()._ = Mock()
    jmp.add_to_api()
    jelapic()._.assert_called_once()



# JelasticAPI

# Configuration variables
api_url = None
api_token = None
hoster_domain = None


from .classes import (  # noqa
    JelasticEnvironment,
    JelasticMountPoint,
    JelasticNode,
    JelasticNodeGroup,
)
from .exceptions import (  # noqa
    JelapiException,
    JelasticAPIException,
    JelasticObjectException,
)

_api_connector = None


def api_connector():
    """
    Get the global jelapi api_connector
    """
    global _api_connector, api_url, api_token
    from .connector import JelasticAPIConnector

    if (
        isinstance(_api_connector, JelasticAPIConnector)
        and _api_connector.is_functional()
        and _api_connector.apiurl == api_url
        and _api_connector.apitoken == api_token
    ):
        # Only return the global one if it is somewhat functional
        return _api_connector

    _api_connector = JelasticAPIConnector(apiurl=api_url, apitoken=api_token)
    return _api_connector



#!/usr/bin/env python3

from setuptools import find_packages, setup

__version__ = "0.1.0rc15"

install_requires = ["httpx[http2]<0.18"]
test_requires = [
    "respx",
    "pytest-cov",
    "factory_boy",
    "faker_enum",
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="jelapi",
    version=__version__,
    author="Didier Raboud",
    author_email="didier.raboud@liip.ch",
    license="GPLv3+",
    description="jelapi: A Jelastic API Python library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/liip/jelapi",
    packages=find_packages(),
    install_requires=install_requires,
    extras_require={
        "test": test_requires,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Topic :: Software Development :: Libraries",
        "Development Status :: 4 - Beta",
    ],
    python_requires=">=3.7",
)



import warnings

from jelapi.classes._volume import _JelasticVolume
from jelapi.classes.environment import JelasticEnvironment
from jelapi.classes.nodegroup import JelasticNodeGroup
from jelapi.factories import JelasticNodeGroupFactory

from .utils import get_standard_env

env = JelasticEnvironment(jelastic_env=get_standard_env())
cp_node_group = JelasticNodeGroupFactory(
    nodeGroupType=JelasticNodeGroup.NodeGroupType.APPLICATION_SERVER
)
cp_node_group.attach_to_environment(env)


def test_JelasticVolume_init():
    """
    Test we can instantiate this without node_group
    """
    _JelasticVolume()


def test_JelasticVolume_node_group_is_deprecated():
    """
    Test we can instantiate this, and take the str representation
    """
    with warnings.catch_warnings(record=True) as warns:
        jv = _JelasticVolume(node_group=cp_node_group)
        assert len(warns) == 1
        str(jv)



import warnings
from copy import deepcopy
from unittest.mock import Mock

import pytest

from jelapi import api_connector as jelapic
from jelapi.classes import JelasticEnvironment, JelasticNode
from jelapi.exceptions import JelasticObjectException
from jelapi.factories import JelasticEnvironmentFactory, JelasticNodeFactory

from .utils import get_standard_node

jelenv = JelasticEnvironmentFactory()
node_group = list(jelenv.nodeGroups.values())[0]


def test_JelasticNode_simple_load():
    """
    JelasticNode can be instantiated as-is
    """
    JelasticNode()


def test_JelasticNode_with_node_group():
    """
    JelasticNode can be instantiated with just node_group
    """
    with warnings.catch_warnings(record=True) as warns:
        JelasticNode(node_group=node_group)
        # There was one deprecationWarning for node_from_env() usage
        assert len(warns) == 1
        assert issubclass(warns[0].category, DeprecationWarning)


def test_JelasticNode_with_node_from_env():
    """
    JelasticNode can be instantiated with just node_from_env
    """
    with warnings.catch_warnings(record=True) as warns:
        JelasticNode(node_from_env=get_standard_node())
        # There was one deprecationWarning for node_from_env() usage
        assert len(warns) == 1
        assert issubclass(warns[0].category, DeprecationWarning)


def test_JelasticNode_update_from_env_dict():
    """
    JelasticNode can get an env dict
    """
    node = JelasticNode()
    assert not node.is_from_api
    node.update_from_env_dict(node_from_env=get_standard_node())
    assert node.is_from_api


def test_JelasticNode_update_from_broken_env_dict():
    """
    JelasticNode cannot be updated from env with missing items
    """
    for musthavekey in [
        "id",
        "name",
        "nodemission",
        "status",
        "type",
        "fixedCloudlets",
        "flexibleCloudlets",
    ]:
        node = get_standard_node()
        del node[musthavekey]

        n = JelasticNode()
        with pytest.raises(KeyError):
            # missing items raises
            n.update_from_env_dict(node)


def test_JelasticNode_update_from_dict_with_unknown_nodeType():
    """
    JelasticNode cannot be instantiated with weird nodeTypes
    """
    node = get_standard_node()
    node["nodeType"] = "this-is-unknown-type"

    n = JelasticNode()
    with pytest.raises(JelasticObjectException):
        n.update_from_env_dict(node)


def test_JelasticNode_factory():
    node = JelasticNodeFactory()
    assert node.is_from_api

    with pytest.raises(JelasticObjectException):
        # envVars cannot be fetched if not from API
        node.envVars


def test_JelasticNode_immutable_data():
    """
    Doesn't differ from API at build
    """
    node = JelasticNodeFactory()
    node._id = 3
    assert str(node) == "JelasticNode id:3"

    with pytest.raises(AttributeError):
        node.id = 4
    with pytest.raises(AttributeError):
        node.envName = "Something else"


def test_JelasticNode_birth_from_api():
    """
    Doesn't differ from API at build
    """
    node = JelasticNodeFactory()
    assert not node.differs_from_api()


def test_JelasticNode_cloudlet_changes_let_differ_from_api():
    """
    Any cloudlet change makes it differ from API
    """
    node = JelasticNodeFactory()
    # Force 1 as coming from API
    node._from_api["fixedCloudlets"] = 1
    node.fixedCloudlets = 8
    assert node.differs_from_api()
    node.fixedCloudlets = 1
    assert not node.differs_from_api()
    node.flexibleCloudlets = 8
    assert node.differs_from_api()


def test_JelasticNode_flexibleCloudlet_reduction_allowance_doesnt_differ_from_api():
    """
    Setting the allowed flag doesn't make the node differ from API by itself
    """
    node = JelasticNodeFactory()
    assert not node.differs_from_api()
    assert not node.allowFlexibleCloudletsReduction

    node.allowFlexibleCloudletsReduction = True
    assert not node.differs_from_api()


def test_JelasticNode_cannot_be_api_updated_without_node_group():
    """
    Test that without envName, no API updates can happen
    """
    node = JelasticNode()
    with pytest.raises(JelasticObjectException):
        node.raise_unless_can_update_to_api()

    # Setting the node_group fixes that
    node.attach_to_node_group(node_group=node_group)
    node.raise_unless_can_update_to_api()


def test_JelasticNode_set_cloudlets():
    """
    Setting any of fixed or flexible cloudlets calls the API once
    """
    jelapic()._ = Mock()
    node = JelasticNodeFactory()
    node.attach_to_node_group(node_group)
    node.fixedCloudlets = 3
    node.save()
    jelapic()._.assert_called_once()


def test_JelasticNode_cannot_reduce_flexibleCloudlets():
    """
    Reducing the flexible cloudlets cannot be done without setting the allowed flag
    """
    node = JelasticNodeFactory()
    node.attach_to_node_group(node_group)
    # 8 came from API
    node._from_api["flexibleCloudlets"] = 8

    node.flexibleCloudlets = 7
    with pytest.raises(JelasticObjectException):
        node.save()

    jelapic()._ = Mock()
    node.allowFlexibleCloudletsReduction = True
    node.save()
    jelapic()._.assert_called_once()


def test_JelasticNode_envVars_works_but_takes_from_node_group():
    """
    Getting the envVars gets us the nodeGroups'
    """
    node_group._envVars = {"TEST": "example.com"}
    node_group._envVars_need_fetching = False

    node = JelasticNodeFactory()
    node.attach_to_node_group(node_group)

    assert "TEST" in node.envVars
    assert node.envVars["TEST"] == "example.com"


def test_JelasticNode_envVars_raises_if_env_is_not_running():
    """
    Saving a faked envVars without fetch will raise
    """
    from jelapi.classes import JelasticEnvironment

    JelStatus = JelasticEnvironment.Status
    for status in JelStatus:
        jelenv_local = deepcopy(jelenv)
        jelenv_local.status = status
        node_group_local = deepcopy(node_group)
        node_group_local._parent = jelenv_local

        node = JelasticNodeFactory()
        node.attach_to_node_group(node_group_local)

        jelapic()._ = Mock(
            return_value={"object": {"VAR": "value"}},
        )
        if status in [
            JelStatus.RUNNING,
            JelStatus.CREATING,
            JelStatus.CLONING,
        ]:
            node.envVars
        else:
            with pytest.raises(JelasticObjectException):
                node.envVars


def test_JelasticNode_links_empty():
    """
    Getting the links works
    """
    node = JelasticNodeFactory()
    assert node.links == []


def test_JelasticNode_links_only_from_inwards():
    """
    Getting the links works
    """
    node = JelasticNode()
    ndict = get_standard_node()
    ndict["customitem"] = {
        "dockerLinks": [
            {"type": "IN", "sourceNodeId": 0},
            {"type": "OUT", "sourceNodeId": 10},
        ]
    }
    node.update_from_env_dict(ndict)
    assert len(node.links) == 1
    assert node.links[0]["sourceNodeId"] == 0


def test_JelasticNode_exec_commands():
    """
    We can launch multiple commands in sequence in nodes
    """
    node = JelasticNode()
    node.attach_to_node_group(node_group)
    node.update_from_env_dict(get_standard_node())

    jelapic()._ = Mock(
        return_value={
            "responses": [
                {
                    "errOut": "",
                    "exitStatus": 0,
                    "nodeid": node.id,
                    "out": "",
                    "result": 0,
                },
                {
                    "errOut": "",
                    "exitStatus": 0,
                    "nodeid": node.id,
                    "out": "example.com",
                    "result": 0,
                },
            ],
            "result": 0,
        },
    )
    assert node.execute_commands(["/bin/true", "echo 'example.com'"])
    jelapic()._.assert_called_once()
    with pytest.raises(TypeError):
        # We can't shortcut the list for a single command
        node.execute_commands("/bin/true")


def test_JelasticNode_exec_command():
    """
    We can launch a single command in a node
    """
    node = JelasticNodeFactory()
    node.attach_to_node_group(node_group)

    jelapic()._ = Mock(
        return_value={
            "responses": [
                {
                    "errOut": "",
                    "exitStatus": 0,
                    "nodeid": node.id,
                    "out": "",
                    "result": 0,
                }
            ],
            "result": 0,
        },
    )
    assert node.execute_command("/bin/true")
    jelapic()._.assert_called_once()
    with pytest.raises(TypeError):
        # We can't extend the single command for a list
        node.execute_command(["/bin/true", "echo 'example.com'"])


def test_JelasticNode_read_file():
    """
    We can gather a single file in a node
    """
    node = JelasticNodeFactory()
    node.attach_to_node_group(node_group)

    jelapic()._ = Mock(
        return_value={
            "body": "Text content",
            "result": 0,
        },
    )
    with pytest.raises(TypeError):
        # We can't read no file
        node.read_file("")

    for status in JelasticEnvironment.Status:
        node.nodeGroup._parent.status = status

        jelapic()._.reset_mock()
        if status != JelasticEnvironment.Status.RUNNING:
            with pytest.raises(JelasticObjectException):
                node.read_file("/tmp/test")
        else:
            body = node.read_file("/tmp/test")
            jelapic()._.assert_called_once()
            assert body == "Text content"
